# -*- coding: utf-8 -*-
"""Landlab component to calculate height above nearest drainage.

@author: D Litwin
"""
from warnings import warn

import numpy as np

from landlab import Component
from landlab.utils import return_array_at_node


class HeightAboveDrainageCalculator(Component):
    """
    Calculate the elevation difference between each node and its nearest
    drainage node in a DEM.

    This component implements the method described by Nobre et al (2011). A
    single direction flow director (D8 or steepest descent) must be run prior
    to HeightAboveDrainageCalculator to supply the flow directions. This component does
    not fill depressions in a DEM, but rather it treats them as drainage nodes.
    For best results, please run one of the available pit filling components
    prior to HeightAboveDrainageCalculator.

    Examples
    --------

    >>> import numpy as np
    >>> from numpy.testing import assert_equal

    >>> from landlab import RasterModelGrid
    >>> from landlab.components import HeightAboveDrainageCalculator, FlowAccumulator

    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.add_zeros("topographic__elevation", at="node")
    >>> mg.set_status_at_node_on_edges(right=mg.BC_NODE_IS_CLOSED, bottom=mg.BC_NODE_IS_FIXED_VALUE, \
                                  left=mg.BC_NODE_IS_CLOSED, top=mg.BC_NODE_IS_CLOSED)
    >>> elev = np.array([[2,1,0,1,2],[3,2,1,2,3],[4,3,2,3,4],[5,4,4,4,5]])
    >>> z[:] = elev.reshape(len(z))
    >>> elev
    array([[2, 1, 0, 1, 2],
       [3, 2, 1, 2, 3],
       [4, 3, 2, 3, 4],
       [5, 4, 4, 4, 5]])

    >>> fa = FlowAccumulator(mg, flow_director="D8")
    >>> fa.run_one_step()

    >>> channel__mask = mg.zeros(at="node")
    >>> channel__mask[[2,7]] = 1
    >>> channel__mask.reshape(elev.shape)
    array([[ 0.,  0.,  1.,  0.,  0.],
       [ 0.,  0.,  1.,  0.,  0.],
       [ 0.,  0.,  0.,  0.,  0.],
       [ 0.,  0.,  0.,  0.,  0.]])

    >>> hd = HeightAboveDrainageCalculator(mg, channel_mask=channel__mask)
    >>> hd.run_one_step()

    >>> mg.at_node["height_above_drainage__elevation"].reshape(elev.shape)  # doctest: +NORMALIZE_WHITESPACE
    array([[ 2.,  0.,  0.,  0.,  0.],
           [ 3.,  2.,  0.,  2.,  3.],
           [ 4.,  2.,  1.,  2.,  4.],
           [ 5.,  4.,  4.,  4.,  5.]])

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    Nobre, A. D., Cuartas, L. A., Hodnett, M., Rennó, C. D., Rodrigues, G.,
    Silveira, A., et al. (2011). Height Above the Nearest Drainage – a
    hydrologically relevant new terrain model. Journal of Hydrology, 404(1),
    13–29. https://doi.org/10.1016/j.jhydrol.2011.03.051

    """

    _name = "HeightAboveDrainageCalculator"

    _unit_agnostic = True

    _info = {
        "channel__mask": {
            "dtype": np.uint8,
            "intent": "in",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "Logical map of at which grid nodes channels are present",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "height_above_drainage__elevation": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Elevation above the nearest channel node",
        },
    }

    def __init__(self, grid, channel_mask="channel__mask"):
        """
        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        channel_mask : field name, array of uint8
            Logical map of nodes where drainage is present
        """
        super().__init__(grid)

        if grid.at_node["flow__receiver_node"].size != grid.size("node"):
            msg = (
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that HeightAboveDrainageCalculator is compatible with "
                "route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )
            raise NotImplementedError(msg)

        self._grid = grid
        self._channel_mask = return_array_at_node(self._grid, channel_mask)
        self._elev = grid.at_node["topographic__elevation"]
        self._receivers = grid.at_node["flow__receiver_node"]
        self._node_order = grid.at_node["flow__upstream_node_order"]

        # height above nearest drainage
        if "height_above_drainage__elevation" in grid.at_node:
            self._hand = grid.at_node["height_above_drainage__elevation"]
        else:
            self._hand = grid.add_zeros(
                "height_above_drainage__elevation", at="node", dtype=float
            )

    @property
    def channel_mask(self):
        return self._channel_mask

    @channel_mask.setter
    def channel_mask(self, new_val):
        self._channel_mask = return_array_at_node(self._grid, new_val)

    def run_one_step(self):

        is_drainage_node = self._channel_mask
        is_drainage_node[self._grid.open_boundary_nodes] = 1

        # check for pits
        self_draining_nodes = np.where(
            self._receivers == np.arange(self._grid.number_of_nodes)
        )
        pits = np.setxor1d(self_draining_nodes, self._grid.boundary_nodes)
        if pits.any():
            warn(
                "Pits detected in the flow directions supplied. "
                "Pits will be treated as drainage nodes."
            )
            is_drainage_node[pits] = 1

        # iterate downstream through stack to find nearest drainage elevation
        nearest_drainage_elev = np.empty(self._elev.shape)
        for n in self._node_order:
            r = self._receivers[n]
            # if not drainage node set drainage elevation to downstream.
            if not is_drainage_node[n]:
                nearest_drainage_elev[n] = nearest_drainage_elev[r]
            else:  # set elevation of drainage to self.
                nearest_drainage_elev[n] = self._elev[n]

        self._hand[:] = self._elev - nearest_drainage_elev



import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.backends.backend_pdf import PdfPages

import landlab


@pytest.mark.slow
def test_imshow_grid():
    rmg = landlab.RasterModelGrid((4, 5))

    pp = PdfPages("test.pdf")

    values = np.arange(rmg.number_of_nodes)
    landlab.plot.imshow_grid(rmg, values, values_at="node", limits=(0, 20))
    pp.savefig()

    plt.clf()
    rmg.status_at_node[7] = rmg.BC_NODE_IS_CLOSED
    values = np.arange(rmg.number_of_cells)
    landlab.plot.imshow_grid(rmg, values, values_at="cell", symmetric_cbar=True)
    pp.savefig()
    pp.close()


def test_imshow_grid_input():
    rmg = landlab.RasterModelGrid((4, 5))
    values = np.arange(rmg.number_of_nodes - 1)
    with pytest.raises(ValueError):
        _ = landlab.plot.imshow_grid(rmg, values, values_at="node", limits=(0, 20))


def test_imshowhs_grid_input():
    rmg = landlab.RasterModelGrid((4, 5))
    values = np.arange(rmg.number_of_nodes - 1)
    with pytest.raises(ValueError):
        _ = landlab.plot.imshowhs_grid(rmg, values, values_at="node", limits=(0, 20))


def test_imshowhs_grid_input_Layer1():
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    values1 = np.arange(mg.number_of_nodes - 1)

    with pytest.raises(ValueError):
        _ = landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            drape1=values1,
            plot_type="Drape1",
            var_name="Soil",
            var_units=r"m",
            grid_units=("m", "m"),
            cmap="terrain",
            thicks_km=False,
            limits=(0, 2),
        )


def test_imshowhs_grid_input_Layer2():
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    values1 = np.arange(mg.number_of_nodes)
    values2 = np.arange(mg.number_of_nodes - 1)

    with pytest.raises(ValueError):
        _ = landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            drape1=values1,
            drape2=values2,
            plot_type="Drape2",
            var_name="Soil",
            var_units=r"m",
            grid_units=("m", "m"),
            cmap="terrain",
            thicks_km=False,
            limits=(0, 2),
        )


def test_imshowhs_grid_1():
    """
    Show DEM draped over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        var_name="Topo",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        symmetric_cbar=True,
        limits=(0, 10),
    )


# %%
def test_imshowhs_grid_2():
    """
    Show DEM draped over the shaded topographic relief with exaggeration
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        var_name="Topo",
        var_units=r"m",
        grid_units=("m", "m"),
        vertical_exa=2,
        thicks_km=True,
        symmetric_cbar=True,
        vmin=0,
        vmax=10,
    )


def test_imshowhs_grid_3():
    """
    Show Hillshade
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        plot_type="Hillshade",
        var_name="Topo",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        plt_contour=True,
        vmax=10,
        vmin=0,
    )


def test_imshowhs_grid_4a():
    """
    Show Drape1 draped over the shaded topographic relief with exaggeration
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    # Show Soil thickness draped over the shaded topographic relief
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["soil__depth"],
        plot_type="Drape1",
        var_name="Soil",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        limits=(0, 2),
    )


def test_imshowhs_grid_4b():
    """
    Show Drape1 draped over the shaded topographic relief with exaggeration
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    # Show Soil thickness draped over the shaded topographic relief
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["soil__depth"],
        plot_type="Drape1",
        var_name="Soil",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        vmin=0,
        vmax=2,
        plt_contour=True,
    )


def test_imshowhs_grid_4c():
    """
    Show Drape1 draped over the shaded topographic relief with exaggeration
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    # Show Soil thickness draped over the shaded topographic relief
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["soil__depth"],
        plot_type="Drape1",
        var_name="Soil",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        symmetric_cbar=True,
    )


# %%


def test_imshowhs_grid_5():
    """
    Show Drape1 draped over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    _ = mg.add_zeros("Layer_1", at="node")
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        plot_type="Drape1",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        limits=(0, 2),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
    )


def test_imshowhs_grid_6a():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    L1 = mg.add_zeros("Layer_1", at="node")
    L2 = mg.add_zeros("Layer_2", at="node")
    L1[:] += 10
    L2[:] += 100
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        limits=(0, 200),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="red",
    )


def test_imshowhs_grid_6b():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax <10
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    L1 = mg.add_zeros("Layer_1", at="node")
    L2 = mg.add_zeros("Layer_2", at="node")
    L1[:] += 10
    L2[:] += 100
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="red",
        vmin=0,
        vmax=9,
    )


def test_imshowhs_grid_6c():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax <100
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    L1 = mg.add_zeros("Layer_1", at="node")
    L2 = mg.add_zeros("Layer_2", at="node")
    L1[:] += 10
    L2[:] += 100
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="red",
        vmin=0,
        vmax=99,
    )


def test_imshowhs_grid_6d():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax <100
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    L1 = mg.add_zeros("Layer_1", at="node")
    L2 = mg.add_zeros("Layer_2", at="node")
    L1[:] += 10
    L2[:] += 100
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="red",
        vmin=0,
        vmax=999,
    )
    # %%


def test_imshowhs_grid_6e():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax <100
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    L1 = mg.add_zeros("Layer_1", at="node")
    L2 = mg.add_zeros("Layer_2", at="node")
    L1[:] += 10
    L2[:] += 100
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1=mg.at_node["Layer_1"],
        drape2=mg.at_node["Layer_2"],
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="red",
        add_double_colorbar=True,
        vmin=0,
        vmax=99999,
    )


# %%


def test_imshowhs_grid_7():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    _ = mg.add_zeros("Layer_1", at="node")
    _ = mg.add_zeros("Layer_2", at="node")
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1="topographic__elevation",
        drape2="soil__depth",
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        limits=(0, 2),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="red",
        thres_drape2=1,
        cmap2=None,
        add_double_colorbar=True,
    )


# %%
def test_imshowhs_grid_8():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax >10<100
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    L1 = mg.add_zeros("Layer_1", at="node")
    L2 = mg.add_zeros("Layer_2", at="node")
    L1[:] += 10
    L2[:] += 100
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1="topographic__elevation",
        drape2="soil__depth",
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        limits=(0, 2),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="red",
        thres_drape2=1,
        cmap2=None,
        add_double_colorbar=True,
        vmin=0,
        vmax=99,
    )


# %%
def test_imshowhs_grid_9():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief, vmax>100
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    _ = mg.add_zeros("Layer_1", at="node")
    _ = mg.add_zeros("Layer_2", at="node")
    _ = landlab.plot.imshowhs_grid(
        mg,
        "topographic__elevation",
        drape1="topographic__elevation",
        drape2="soil__depth",
        plot_type="Drape2",
        var_name="Layer 1",
        var_units=r"m",
        grid_units=("m", "m"),
        cmap="terrain",
        thicks_km=False,
        limits=(0, 2),
        colorbar_label_y=-55,
        add_label_bbox=True,
        thres_drape1=0.001,
        color_for_closed="red",
        thres_drape2=1,
        cmap2=None,
        add_double_colorbar=True,
        vmin=0,
        vmax=99999,
    )
    with pytest.raises(ValueError):
        _ = landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            plot_type="Oops",
        )


def test_imshowhs_grid_10():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    _ = mg.add_zeros("Layer_1", at="node")
    _ = mg.add_zeros("Layer_2", at="node")
    with pytest.raises(ValueError):
        _ = landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            drape1=mg.at_node["Layer_1"],
            plot_type="Drape2",
            var_name="Layer 1",
            var_units=r"m",
            grid_units=("m", "m"),
            cmap="terrain",
            thicks_km=False,
            limits=(0, 2),
            colorbar_label_y=-55,
            add_label_bbox=True,
            thres_drape1=0.001,
        )


def test_imshowhs_grid_11():
    """
    Show Layer 1 and Layer 2 over the shaded topographic relief
    """
    # %%
    mg = landlab.RasterModelGrid((4, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    _ = mg.add_zeros("Layer_1", at="node")
    _ = mg.add_zeros("Layer_2", at="node")
    with pytest.raises(ValueError):
        _ = landlab.plot.imshowhs_grid(
            mg,
            "topographic__elevation",
            plot_type="Drape1",
            var_name="Layer 1",
            var_units=r"m",
            grid_units=("m", "m"),
            cmap="terrain",
            thicks_km=False,
            limits=(0, 2),
            colorbar_label_y=-55,
            add_label_bbox=True,
            thres_drape1=0.001,
        )


def test_hex_mfd():
    """
    Currently no support for hex
    """
    # %%
    mg = landlab.HexModelGrid((5, 3))
    _ = mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(NotImplementedError):
        _ = landlab.plot.imshowhs_grid(mg, "topographic__elevation")


# %%
def test_at_cell():
    """
    Currently no support for at cell
    """
    # %%
    mg = landlab.HexModelGrid((5, 3))
    _ = mg.add_field("topographic__elevation", np.zeros((7,)), at="cell")
    with pytest.raises(NotImplementedError):
        _ = landlab.plot.imshowhs_grid(mg, "topographic__elevation", at="cell")


# %%
def test_at_other():
    """
    Currently no support for non at node valley locations
    """
    # %%
    mg = landlab.HexModelGrid((5, 3))
    _ = mg.add_field("topographic__elevation", np.zeros((24,)), at="corner")
    with pytest.raises(TypeError):
        _ = landlab.plot.imshowhs_grid(mg, "topographic__elevation", at="corner")



import pathlib

import pytest

from landlab import ExampleData


@pytest.mark.parametrize("case", (None, "", "methow"))
def test_fetch(tmpdir, case):
    if case is None:
        data = ExampleData("io/shapefile")
    else:
        data = ExampleData("io/shapefile", case=case)
    expected = set(data)
    with tmpdir.as_cwd():
        data.fetch()

        fetched = set(p.name for p in pathlib.Path(".").iterdir())
        assert expected == fetched


def test_fetch_all_files(tmpdir):
    with tmpdir.as_cwd():
        ExampleData("io/shapefile", case="methow").fetch()
        fetched_with_case = set(p.name for p in pathlib.Path(".").iterdir())

        ExampleData("io/shapefile").fetch()
        fetched_without_case = set(p.name for p in pathlib.Path("methow").iterdir())

        assert fetched_with_case == fetched_without_case


def test_fetch_no_overwrite(tmpdir):
    data = ExampleData("io/shapefile")
    with tmpdir.as_cwd():
        pathlib.Path(list(data).pop()).touch()
        with pytest.raises(FileExistsError):
            data.fetch()



#! /usr/bin/env python
"""Defines the base component class from which Landlab components inherit.

Base component class methods
++++++++++++++++++++++++++++

.. autosummary::

    ~landlab.core.model_component.Component.name
    ~landlab.core.model_component.Component.from_path
    ~landlab.core.model_component.Component.unit_agnostic
    ~landlab.core.model_component.Component.units
    ~landlab.core.model_component.Component.definitions
    ~landlab.core.model_component.Component.input_var_names
    ~landlab.core.model_component.Component.output_var_names
    ~landlab.core.model_component.Component.optional_var_names
    ~landlab.core.model_component.Component.var_type
    ~landlab.core.model_component.Component.var_units
    ~landlab.core.model_component.Component.var_definition
    ~landlab.core.model_component.Component.var_mapping
    ~landlab.core.model_component.Component.var_loc
    ~landlab.core.model_component.Component.var_help
    ~landlab.core.model_component.Component.initialize_output_fields
    ~landlab.core.model_component.Component.initialize_optional_output_fields
    ~landlab.core.model_component.Component.shape
    ~landlab.core.model_component.Component.grid
    ~landlab.core.model_component.Component.coords
"""

import os
import textwrap

import numpy as np

from .. import registry
from ..field import FieldError
from .model_parameter_loader import load_params

_VAR_HELP_MESSAGE = """
name: {name}
description:
{desc}
units: {units}
unit agnostic: {unit_agnostic}
at: {loc}
intent: {intent}
"""


class classproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class Component:
    """Base component class from which Landlab components inherit."""

    _info = {}
    _name = None
    _cite_as = ""
    _unit_agnostic = None

    def __new__(cls, *args, **kwds):
        registry.add(cls)
        return object.__new__(cls)

    def __init__(self, grid):
        self._grid = grid
        self._current_time = None
        # ensure that required input fields exist
        for name in self._info.keys():
            at = self._info[name]["mapping"]
            optional = self._info[name]["optional"]
            in_true = "in" in self._info[name]["intent"]
            if (in_true) and (not optional):
                # if required input, verify that it exists.
                if name not in self._grid[at]:
                    raise FieldError(
                        "{component} is missing required input field: {name} at {at}".format(
                            component=self._name, name=name, at=at
                        )
                    )

                # if required input exists, check dtype.
                field = self._grid[at][name]
                dtype = self._info[name]["dtype"]

                if field.dtype != dtype:
                    raise FieldError(
                        "{component} required input variable: {name} at {at} has incorrect dtype. dtype must be {dtype} and is {actual}".format(
                            component=self._name,
                            name=name,
                            at=at,
                            dtype=dtype,
                            actual=field.dtype,
                        )
                    )

            # if optional input exists, check dtype
            if in_true and optional:

                if name in self._grid[at]:
                    field = self._grid[at][name]
                    dtype = self._info[name]["dtype"]

                    if field.dtype != dtype:
                        raise FieldError(
                            "{component} optional input variable: {name} at {at} has incorrect dtype. dtype must be {dtype} and is {actual}".format(
                                component=self._name,
                                name=name,
                                at=at,
                                dtype=dtype,
                                actual=field.dtype,
                            )
                        )

    @classmethod
    def from_path(cls, grid, path):
        """Create a component from an input file.

        Parameters
        ----------
        grid : ModelGrid
            A landlab grid.
        path : str or file_like
            Path to a parameter file, contents of a parameter file, or
            a file-like object.

        Returns
        -------
        Component
            A newly-created component.
        """
        if os.path.isfile(path):
            with open(path, "r") as fp:
                params = load_params(fp)
        else:
            params = load_params(path)
        return cls(grid, **params)

    @classproperty
    @classmethod
    def cite_as(cls):
        """Citation information for component.

        Return required software citation, if any. An empty string indicates
        that no citations other than the standard Landlab package citations are
        needed for the component.

        Citations are provided in BibTeX format.

        Returns
        -------
        cite_as
        """
        return cls._cite_as

    @property
    def current_time(self):
        """Current time.

        Some components may keep track of the current time. In this case, the
        ``current_time`` attribute is incremented. Otherwise it is set to None.

        Returns
        -------
        current_time
        """
        return self._current_time

    @current_time.setter
    def current_time(self, new_time):
        if self._current_time is not None:
            assert new_time > self._current_time
        self._current_time = new_time

    @classproperty
    @classmethod
    def input_var_names(cls):
        """Names of fields that are used by the component.

        Returns
        -------
        tuple of str
            Tuple of field names.
        """
        input_var_names = [
            name
            for name in cls._info.keys()
            if (not cls._info[name]["optional"]) and ("in" in cls._info[name]["intent"])
        ]
        return tuple(sorted(input_var_names))

    @classproperty
    @classmethod
    def output_var_names(cls):
        """Names of fields that are provided by the component.

        Returns
        -------
        tuple of str
            Tuple of field names.
        """
        output_var_names = [
            name
            for name in cls._info.keys()
            if (not cls._info[name]["optional"])
            and ("out" in cls._info[name]["intent"])
        ]
        return tuple(sorted(output_var_names))

    @classproperty
    @classmethod
    def optional_var_names(cls):
        """Names of fields that are optionally provided by the component, if
        any.

        Returns
        -------
        tuple of str
            Tuple of field names.
        """
        optional_var_names = [
            name for name in cls._info.keys() if cls._info[name]["optional"]
        ]
        return tuple(sorted(optional_var_names))

    @classmethod
    def var_type(cls, name):
        """Returns the dtype of a field (float, int, bool, str...).

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        dtype
            The dtype of the field.
        """
        return cls._info[name]["dtype"]

    @classproperty
    @classmethod
    def name(cls):
        """Name of the component.

        Returns
        -------
        str
            Component name.
        """
        return cls._name

    @classproperty
    @classmethod
    def unit_agnostic(cls):
        """Whether the component is unit agnostic.

        If True, then the component is unit agnostic. Under this condition a
        user must still provide consistent units across all input arguments,
        keyword arguments, and fields. However, when ``unit_agnostic`` is True
        the units specified can be interpreted as dimensions.

        When False, then the component requires inputs in the specified units.

        Returns
        -------
        bool

        """
        return cls._unit_agnostic

    @classproperty
    @classmethod
    def units(cls):
        """Get the units for all field values.

        Returns
        -------
        tuple or str
            Units for each field.
        """
        return tuple(
            sorted([(name, cls._info[name]["units"]) for name in cls._info.keys()])
        )

    @classmethod
    def var_units(cls, name):
        """Get the units of a particular field.

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        str
            Units for the given field.
        """
        return cls._info[name]["units"]

    @classproperty
    @classmethod
    def definitions(cls):
        """Get a description of each field.

        Returns
        -------
        tuple of (*name*, *description*)
            A description of each field.
        """
        return tuple(
            sorted([(name, cls._info[name]["doc"]) for name in cls._info.keys()])
        )

    @classmethod
    def var_definition(cls, name):
        """Get a description of a particular field.

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        tuple of (*name*, *description*)
            A description of each field.
        """
        return cls._info[name]["doc"]

    @classmethod
    def var_help(cls, name):
        """Print a help message for a particular field.

        Parameters
        ----------
        name : str
            A field name.
        """
        desc = os.linesep.join(
            textwrap.wrap(
                cls._info[name]["doc"], initial_indent="  ", subsequent_indent="  "
            )
        )
        units = cls._info[name]["units"]
        loc = cls._info[name]["mapping"]
        intent = cls._info[name]["intent"]

        help = _VAR_HELP_MESSAGE.format(
            name=name,
            desc=desc,
            units=units,
            loc=loc,
            intent=intent,
            unit_agnostic=cls._unit_agnostic,
        )

        print(help.strip())

    @classproperty
    @classmethod
    def var_mapping(cls):
        """Location where variables are defined.

        Returns
        -------
        tuple of (name, location)
            Tuple of variable name and location ('node', 'link', etc.) pairs.
        """
        return tuple(
            sorted([(name, cls._info[name]["mapping"]) for name in cls._info.keys()])
        )

    @classmethod
    def var_loc(cls, name):
        """Location where a particular variable is defined.

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        str
            The location ('node', 'link', etc.) where a variable is defined.
        """
        return cls._info[name]["mapping"]

    def initialize_output_fields(self, values_per_element=None):
        """Create fields for a component based on its input and output var
        names.

        This method will create new fields (without overwrite) for any fields
        output by, but not supplied to, the component. New fields are
        initialized to zero. Ignores optional fields. New fields are created as
        arrays of floats, unless the component specifies the variable type.

        Parameters
        ----------
        values_per_element: int (optional)
            On occasion, it is necessary to create a field that is of size
            (n_grid_elements, values_per_element) instead of the default size
            (n_grid_elements,). Use this keyword argument to acomplish this
            task.
        """
        for name in self._info.keys():
            at = self._info[name]["mapping"]
            optional = self._info[name]["optional"]
            out_true = "out" in self._info[name]["intent"]
            if (out_true) and (not optional) and (name not in self._grid[at]):

                type_in = self.var_type(name)
                num_elements = self._grid.size(at)

                if values_per_element is None:
                    size = num_elements
                else:
                    size = (num_elements, values_per_element)

                init_vals = np.zeros(size, dtype=type_in)
                units_in = self.var_units(name)

                self.grid.add_field(name, init_vals, at=at, units=units_in, copy=False)

    def initialize_optional_output_fields(self):
        """Create fields for a component based on its optional field outputs,
        if declared in _optional_var_names.

        This method will create new fields (without overwrite) for any
        fields output by the component as optional. New fields are
        initialized to zero. New fields are created as arrays of floats,
        unless the component also contains the specifying property
        _var_type.
        """

        for name in self._info.keys():
            at = self._info[name]["mapping"]
            optional = self._info[name]["optional"]
            out_true = "out" in self._info[name]["intent"]
            if (out_true) and (optional) and (name not in self._grid[at]):

                type_in = self.var_type(name)
                init_vals = self.grid.zeros(at, dtype=type_in)
                units_in = self.var_units(name)

                self.grid.add_field(name, init_vals, at=at, units=units_in, copy=False)

    @property
    def shape(self):
        """Return the grid shape attached to the component, if defined."""
        return self.grid._shape

    @property
    def grid(self):
        """Return the grid attached to the component."""
        return self._grid

    @property
    def coords(self):
        """Return the coordinates of nodes on grid attached to the
        component."""
        return (self.grid.node_x, self.grid.node_y)



"""Landlab component that simulates detachment-limited river erosion.

This component calculates changes in elevation in response to
vertical incision.
"""

import numpy as np

from landlab import Component


class DetachmentLtdErosion(Component):
    """Simulate detachment limited sediment transport.

    Landlab component that simulates detachment limited sediment transport is more
    general than the stream power component. Doesn't require the upstream node
    order, links to flow receiver and flow receiver fields. Instead, takes in
    the discharge values on NODES calculated by the OverlandFlow class and
    erodes the landscape in response to the output discharge.

    As of right now, this component relies on the OverlandFlow component
    for stability. There are no stability criteria implemented in this class.
    To ensure model stability, use StreamPowerEroder or FastscapeEroder
    components instead.

    .. codeauthor:: Jordan Adams

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import DetachmentLtdErosion

    Create a grid on which to calculate detachment ltd sediment transport.

    >>> grid = RasterModelGrid((4, 5))

    The grid will need some data to provide the detachment limited sediment
    transport component. To check the names of the fields that provide input to
    the detachment ltd transport component, use the *input_var_names* class
    property.

    Create fields of data for each of these input variables.

    >>> grid.at_node['topographic__elevation'] = np.array([
    ...     0., 0., 0., 0., 0.,
    ...     1., 1., 1., 1., 1.,
    ...     2., 2., 2., 2., 2.,
    ...     3., 3., 3., 3., 3.])

    Using the set topography, now we will calculate slopes on all nodes.


    >>> grid.at_node['topographic__slope'] = np.array([
    ...     -0.        , -0.        , -0.        , -0.        , -0,
    ...      0.70710678,  1.        ,  1.        ,  1.        ,  0.70710678,
    ...      0.70710678,  1.        ,  1.        ,  1.        ,  0.70710678,
    ...      0.70710678,  1.        ,  1.        ,  1.        ,  0.70710678])


    Now we will arbitrarily add water discharge to each node for simplicity.
    >>> grid.at_node['surface_water__discharge'] = np.array([
    ...     30., 30., 30., 30., 30.,
    ...     20., 20., 20., 20., 20.,
    ...     10., 10., 10., 10., 10.,
    ...      5., 5., 5., 5., 5.])

    Instantiate the `DetachmentLtdErosion` component to work on this grid, and
    run it. In this simple case, we need to pass it a time step ('dt')

    >>> dt = 10.0
    >>> dle = DetachmentLtdErosion(grid)
    >>> dle.run_one_step(dt=dt)

    After calculating the erosion rate, the elevation field is updated in the
    grid. Use the *output_var_names* property to see the names of the fields that
    have been changed.

    >>> dle.output_var_names
    ('topographic__elevation',)

    The `topographic__elevation` field is defined at nodes.

    >>> dle.var_loc('topographic__elevation')
    'node'


    Now we test to see how the topography changed as a function of the erosion
    rate.

    >>> grid.at_node['topographic__elevation'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
            0.99936754,  0.99910557,  0.99910557,  0.99910557,  0.99936754,
            1.99955279,  1.99936754,  1.99936754,  1.99936754,  1.99955279,
            2.99968377,  2.99955279,  2.99955279,  2.99955279,  2.99968377])

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    Howard, A. (1994). A detachment-limited model of drainage basin evolution. Water
    Resources Research  30(7), 2261-2285. https://dx.doi.org/10.1029/94wr00757
    """

    _name = "DetachmentLtdErosion"

    _unit_agnostic = True

    _info = {
        "surface_water__discharge": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m**3/s",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__slope": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "gradient of the ground surface",
        },
    }

    def __init__(
        self,
        grid,
        K_sp=0.00002,
        m_sp=0.5,
        n_sp=1.0,
        uplift_rate=0.0,
        entrainment_threshold=0.0,
        slope="topographic__slope",
    ):
        """Calculate detachment limited erosion rate on nodes.

        Landlab component that generalizes the detachment limited erosion
        equation, primarily to be coupled to the the Landlab OverlandFlow
        component.

        This component adjusts topographic elevation.

        Parameters
        ----------
        grid : RasterModelGrid
            A landlab grid.
        K_sp : float, optional
            K in the stream power equation (units vary with other parameters -
            if used with the de Almeida equation it is paramount to make sure
            the time component is set to *seconds*, not *years*!)
        m_sp : float, optional
            Stream power exponent, power on discharge
        n_sp : float, optional
            Stream power exponent, power on slope
        uplift_rate : float, optional
            changes in topographic elevation due to tectonic uplift
        entrainment_threshold : float, optional
            threshold for sediment movement
        slope : str
            Field name of an at-node field that contains the slope.
        """
        super().__init__(grid)

        assert slope in grid.at_node

        self._K = K_sp
        self._m = m_sp
        self._n = n_sp

        self._I = self._grid.zeros(at="node")  # noqa: E741
        self._uplift_rate = uplift_rate
        self._entrainment_threshold = entrainment_threshold

        self._dzdt = self._grid.zeros(at="node")

    def run_one_step(self, dt):
        """Erode into grid topography.

        For one time step, this erodes into the grid topography using
        the water discharge and topographic slope.

        The grid field 'topographic__elevation' is altered each time step.

        Parameters
        ----------
        dt : float
            Time step.
        """

        S = self._grid.at_node["topographic__slope"]
        Q = self._grid.at_node["surface_water__discharge"]

        Q_to_m = np.power(Q, self._m)

        S_to_n = np.power(S, self._n)

        self._I = (
            self._K * Q_to_m * S_to_n
        ) - self._entrainment_threshold  # noqa: E741

        self._I[self._I < 0.0] = 0.0

        self._dz = (self._uplift_rate - self._I) * dt

        self._grid["node"]["topographic__elevation"] += self._dz



"""generate_overland_flow.py.

This component simulates overland flow using
the 2-D numerical model of shallow-water flow
over topography using the Bates et al. (2010)
algorithm for storage-cell inundation modeling.

Written by Jordan Adams, based on code written by Greg Tucker.

Last updated: April 21, 2016
"""
import numpy as np
import scipy.constants

from landlab import Component


class OverlandFlowBates(Component):
    u"""Simulate overland flow using Bates et al. (2010).

    Landlab component that simulates overland flow using the Bates et al.,
    (2010) approximations of the 1D shallow water equations to be used for 2D
    flood inundation modeling.

    This component calculates discharge, depth and shear stress after some
    precipitation event across any raster grid. Default input file is named
    "overland_flow_input.txt' and is contained in the
    landlab.components.overland_flow folder.

    Parameters
    ----------
    grid : RasterGridModel
        A grid.
    input_file : str
        Contains necessary and optional inputs. If not given, default input
        file is used.

        -  Manning's n is *required*.
        -  Storm duration is needed *if* rainfall_duration is not passed in the
           initialization
        -  Rainfall intensity is needed *if* rainfall_intensity is not passed
           in the initialization
        -  Model run time can be provided in initialization. If not it is set
           to the storm duration

    h_init : float, optional
        Some initial depth in the channels. Default = 0.001 m
    g : float, optional
        Gravitational acceleration, :math:`m / s^2`
    alpha : float, optional
        Non-dimensional time step factor from Bates et al., (2010)
    rho : integer, optional
        Density of water, :math:`kg / m^3`
    ten_thirds : float, optional
        Precalculated value of :math:`10 / 3` which is used in the
        implicit shallow water equation.

    Examples
    --------
    >>> DEM_name = 'DEM_name.asc'
    >>> (rg, z) = read_esri_ascii(DEM_name) # doctest: +SKIP
    >>> of = OverlandFlowBates(rg) # doctest: +SKIP

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    Bates, P., Horritt, M., Fewtrell, T. (2010). A simple inertial formulation
    of the shallow water equations for efficient two-dimensional flood
    inundation modelling Journal of Hydrology  387(1-2), 33-45.
    https://dx.doi.org/10.1016/j.jhydrol.2010.03.027

    """

    _name = "OverlandFlowBates"

    _unit_agnostic = False

    _info = {
        "surface_water__depth": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of water on the surface",
        },
        "surface_water__discharge": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m3/s",
            "mapping": "link",
            "doc": "Volumetric discharge of surface water",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    def __init__(
        self,
        grid,
        h_init=0.00001,
        alpha=0.7,
        mannings_n=0.03,
        g=scipy.constants.g,
        rainfall_intensity=0.0,
    ):
        super().__init__(grid)

        # First we copy our grid

        self._h_init = h_init
        self._alpha = alpha
        self._mannings_n = mannings_n
        self._g = g
        self._rainfall_intensity = rainfall_intensity

        # Now setting up fields at the links...
        # For water discharge

        self._surface_water__discharge = grid.add_zeros(
            "surface_water__discharge",
            at="link",
            units=self._info["surface_water__discharge"]["units"],
        )

        # Pre-calculated values included for speed.
        self._ten_thirds = 10.0 / 3.0
        self._mannings_n_squared = self._mannings_n * self._mannings_n

        # Start time of simulation is at 1.0 s
        self._elapsed_time = 1.0

        # Assigning a class variable to the water depth field and adding the
        # initial thin water depth
        self._h = self._grid["node"]["surface_water__depth"] = (
            self._grid["node"]["surface_water__depth"] + self._h_init
        )

        # Assigning a class variable to the water discharge field.
        self._q = self._grid["link"]["surface_water__discharge"]

        # Assiging a class variable to the elevation field.
        self._z = self._grid.at_node["topographic__elevation"]

    @property
    def surface_water__discharge(self):
        """The discharge of water on active links."""
        return self._surface_water__discharge

    @property
    def h(self):
        """The depth of water at each node."""
        return self._h

    @property
    def dt(self):
        """dt: component timestep."""
        return self._dt

    @dt.setter
    def dt(self, dt):
        assert dt > 0
        self._dt = dt

    def calc_time_step(self):

        # Adaptive time stepper from Bates et al., 2010 and de Almeida et al.,
        # 2012
        self._dt = (
            self._alpha
            * self._grid.dx
            / np.sqrt(self._g * np.amax(self._grid.at_node["surface_water__depth"]))
        )

        return self._dt

    def overland_flow(self, dt=None):
        """For one time step, this generates 'overland flow' across a given
        grid by calculating discharge at each node.

        Using the depth slope product, shear stress is calculated at every
        node.

        Outputs water depth, discharge and shear stress values through time at
        every point in the input grid.


        Parameters
        ----------
        grid : RasterModelGrid
            A grid.
        dt : float, optional
            Time step. Either set when called or the component will do it for
            you.
        """

        # If no dt is provided, one will be calculated using
        # self._gear_time_step()
        if dt is None:
            self.calc_time_step()

        # In case another component has added data to the fields, we just reset
        # our water depths, topographic elevations and water discharge
        # variables to the fields.
        # self._h = self._grid['node']['surface_water__depth']
        self._z = self._grid["node"]["topographic__elevation"]
        self._q = self._grid["link"]["surface_water__discharge"]

        # Here we identify the core nodes and active link ids for later use.
        self._core_nodes = self._grid.core_nodes
        self._active_links = self._grid.active_links

        # Per Bates et al., 2010, this solution needs to find the difference
        # between the highest water surface in the two cells and the highest
        # bed elevation
        zmax = self._grid.map_max_of_link_nodes_to_link(self._z)
        w = self._h + self._z
        wmax = self._grid.map_max_of_link_nodes_to_link(w)
        hflow = wmax[self._grid.active_links] - zmax[self._grid.active_links]

        # Now we calculate the slope of the water surface elevation at active
        # links
        water_surface_slope = self._grid.calc_grad_at_link(w)[self._grid.active_links]

        # Here we calculate discharge at all active links using Eq. 11 from
        # Bates et al., 2010
        self._q[self._active_links] = (
            self._q[self._active_links]
            - self._g * hflow * self._dt * water_surface_slope
        ) / (
            1.0
            + self._g
            * hflow
            * self._dt
            * self._mannings_n_squared
            * abs(self._q[self._active_links])
            / hflow ** self._ten_thirds
        )

        # Update our water depths
        dhdt = self._rainfall_intensity - self._grid.calc_flux_div_at_node(self._q)

        self._h[self._core_nodes] = (
            self._h[self._core_nodes] + dhdt[self._core_nodes] * self._dt
        )

        # And reset our field values with the newest water depth and discharge.
        self._grid.at_node["surface_water__depth"] = self._h
        self._grid.at_link["surface_water__discharge"] = self._q



# coding: utf8
# ! /usr/env/python
"""trickle_down_profiler.py component to create channel profiles."""
from collections import OrderedDict

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cm

from landlab.components.profiler.base_profiler import _BaseProfiler
from landlab.core.utils import as_id_array
from landlab.utils.flow__distance import calculate_flow__distance


class TrickleDownProfiler(_BaseProfiler):
    """Extract and a profile from one or more node IDs to their downstream termini.

    The TrickleDownProfiler extracts channel networks from a landlab grid.
    Unlike the ChannelProfiler which starts at one or more watershed outlets
    and works upstream until it reaches the end of the channel (based on
    a specified threshold, such as drainage area) the TrickleDownProfiler
    starts at a *starting node* and works its way downhill until it reaches
    an outlet or sink.

    In order to follow the channel network, the flow connectivity across the
    grid must already be identified. This is typically done with the
    FlowAccumulator component. However, this component does not require that the
    FlowAccumulator was used. Instead it expects that the following at-node
    grid fields will be present:
    ::

        'flow__receiver_node'
        'flow__link_to_receiver_node'

    The TrickleDownProfiler can work on grids that have used route-to-one or
    route-to-multiple flow directing.

    To understand how this component works it is useful to define the following
    terms: *outlet*, *starting node*, and *segment*.

    Consider the following grid with 10 columns and 7 rows. ``@`` represents
    the *starting node*, ``.`` represents the nodes downstream, and the
    watershed outlet node is indicated by ``o``.

    In this and the following examples, we will use only D4 connectivity. The
    ChannelProfiler, however, knows nothing of connectivity other than what is
    implied by the two required grid fields.
    ::

        X X X X X X X X X X
        X X X X X X X X X X
        X X X X X X . . @ X
        X X X X X X . X X X
        X X X . . . . X X X
        X X X . X X X X X X
        X X X o X X X X X X

    For each starting node, the TrickleDownProfiler follows the network
    downstream until it reaches the outlet or sink. One or more starting nodes
    can be used, depending on a user's needs.

    The node IDs and distances upstream of the channel network are stored in
    ``data_structure``. It is a dictionary with keys indicating the starting
    node.

    For each starting node, the value in the ``data_structure`` is itself
    a dictionary with keys that are a segment ID tuple of the
    ``(dowstream, upstream)`` nodes IDs of each channel segment.

    For our simple example, these are the node IDs:
    ::

            X  X  X  X  X  X  X  X  X  X
            X  X  X  X  X  X  X  X  X  X
            X  X  X  X  X  X 46 47 48  X
            X  X  X  X  X  X 36  X  X  X
            X  X  X 23 24 25 26  X  X  X
            X  X  X 13  X  X  X  X  X  X
            X  X  X  3  X  X  X  X  X  X

    The starting node is 48 and the outlet node is 3.

    The value associated with the segment ID tuple ``(3, 48)`` is itself a
    dictionary. It has three key-value pairs. First, ``"ids"`` contains a list
    of the segment node ids ordered from downstream to upstream. It includes
    the endpoints. Second, ``"distances"`` contains a list of distances
    upstream that mirrors the list in ``"ids"``. Finally, ``"color"`` is an
    RGBA tuple indicating the color for the segment.

    By default a unique color will be assigned to each starting node. To change
    the color, a user can change values stored in ``data_structure``.
    Additionally, a ``cmap`` keyword argument can provide some user control
    over the color at the instantiation of the component.

    For example with a starting node of 48, the data structure will look as
    follows:

    .. code-block:: python

        {48: {
            (3, 48) : {
                "ids": [3, 13, 23, 24, 25, 26, 36, 46, 47, 48],
                "distances": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
                "color": (1, 0, 1, 1),
                }
            }
        }

    Note that the distances upstream are relative to the outlet.

    Next consider an example with two starting nodes, each noted with an ``@``.
    ::

        X X X X X X X X X X
        X X X X X X @ X X X
        X X X X @ X . X X X
        o . . . . X . X X X
        X X X X X X . X X X
        X X X X X X . X X X
        X X X X X X . . . X
        X X X X X X X X o X

    And the following node IDs.
    ::

        X   X   X   X   X   X   X   X   X   X
        X   X   X   X   X   X  66   X   X   X
        X   X   X   X  54   X  56   X   X   X
       40  41  42  43  44   X  46   X   X   X
        X   X   X   X   X   X  36   X   X   X
        X   X   X   X   X   X  26   X   X   X
        X   X   X   X   X   X  16  17  18   X
        X   X   X   X   X   X   X   X   8   X

    With our starting nodes of 54 and 66 our data structure will look like.

    .. code-block:: python

        {54: {
            (40, 54) : {
                "ids": [40, 41, 42, 43, 44, 54],
                "distances": [0, 1, 3, 4, 5, 6],
                "color": [ 0.27,  0.  ,  0.33,  1.  ],
                },
            }
        66: {
            (8, 66) : {
                "ids": [8, 18, 17, 16, 26, 36, 46, 56, 66],
                "distances": [0, 1, 2, 3, 4, 5, 6, 7, 8],
                "color": [ 0.13,  0.57,  0.55,  1.  ],
                },
        }

    Examples
    --------

    Start by importing necessary modules

    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator, TrickleDownProfiler

    Create the second example grid we showed above. Note that in order to do
    this we need to enter the elevations starting from the lower left so the
    elevation order may seem upside-down. In addition, in this example,
    elevation is only provided along the profiles. The third line of code below
    sets all nodes with a value of zero to closed, such that these nodes are
    igored.
    >>> z = np.array([ 0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
    ...                0,  0,  0,  0,  0,  0,  4,  3,  2,  0,
    ...                0,  0,  0,  8,  7,  6,  5,  0,  0,  0,
    ...                0,  0,  0,  0,  0,  0,  6,  0,  0,  0,
    ...                1,  3,  4,  5,  6,  0,  7,  0,  0,  0,
    ...                0,  4,  0,  0,  7,  0,  8,  0,  0,  0,
    ...                0,  5,  6,  0,  0,  0,  9,  0,  0,  0,
    ...                0,  0,  0,  0,  0,  0,  0,  0,  0,  0,], dtype=float)

    >>> mg = RasterModelGrid((8, 10))
    >>> z = mg.add_field("topographic__elevation", z, at="node")
    >>> mg.set_nodata_nodes_to_closed(z, 0)
    >>> fa = FlowAccumulator(mg, flow_director='D4')
    >>> fa.run_one_step()
    >>> fa.node_drainage_area.reshape(mg.shape)
    array([[  0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,  11.,   0.],
           [  0.,   0.,   0.,   0.,   0.,   0.,   9.,  10.,  11.,   0.],
           [  0.,   0.,   0.,   1.,   2.,   3.,   8.,   0.,   0.,   0.],
           [  0.,   0.,   0.,   0.,   0.,   0.,   4.,   0.,   0.,   0.],
           [  8.,   8.,   4.,   3.,   2.,   0.,   3.,   0.,   0.,   0.],
           [  0.,   3.,   0.,   0.,   1.,   0.,   2.,   0.,   0.,   0.],
           [  0.,   2.,   1.,   0.,   0.,   0.,   1.,   0.,   0.,   0.],
           [  0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.,   0.]])

    >>> profiler = TrickleDownProfiler(
    ...     mg,
    ...     starting_nodes=[54, 66])
    >>> profiler.run_one_step()

    The keys of the property ``data_structure`` are the IDs of the two outlet
    nodes.

    >>> profiler.data_structure.keys()
    odict_keys([54, 66])

    Within the data structure, the value at key 54, is a dictionary of the
    one segment, each specified by a ``(dowstream, upstream)`` tuple:

    >>> profiler.data_structure[54].keys()
    dict_keys([(40, 54)])

    The value of the segment between nodes 40 and 54 has the following
    components:

    >>> profiler.data_structure[54][(40, 54)]["ids"]
    array([40, 41, 42, 43, 44, 54])
    >>> profiler.data_structure[54][(40, 54)]["distances"]
    array([ 0.,  1.,  2.,  3.,  4.,  5.])
    >>> np.round(profiler.data_structure[54][(40, 54)]["color"], decimals=2)
    array([ 0.27,  0.  ,  0.33,  1.  ])

    The rest of the ``profile_structure`` encodes information about the second
    profile which starts at node 66.

    >>> profiler.data_structure[66].keys()
    dict_keys([(8, 66)])

    >>> profiler.data_structure[66][(8, 66)]["ids"]
    array([ 8, 18, 17, 16, 26, 36, 46, 56, 66])
    >>> profiler.data_structure[66][(8, 66)]["distances"]
    array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.])
    >>> np.round(profiler.data_structure[66][(8, 66)]["color"], decimals=2)
    array([ 0.13,  0.57,  0.55,  1.  ])


    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    None Listed

    """

    _name = "TrickleDownProfiler"

    _unit_agnostic = True

    _info = {
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
    }

    def __init__(
        self,
        grid,
        starting_nodes=None,
        cmap="viridis",
    ):
        """
        Parameters
        ----------
        grid : Landlab Model Grid instance
        starting_nodes : iterable
        cmap : str, optional
            A valid matplotlib cmap string. Default is "viridis".

        """
        super().__init__(grid)

        self._cmap = plt.get_cmap(cmap)

        self._flow_receiver = grid.at_node["flow__receiver_node"]
        self._starting_nodes = starting_nodes

    @property
    def data_structure(self):
        """OrderedDict defining the trickle down network.

        The IDs and upstream distance of the channel network nodes are stored
        in ``data_structure``. It is a dictionary with keys of the outlet node
        ID.

        For each starting node, the value in the ``data_structure`` is
        itself a dictionary with keys that are a segment ID tuple of the
        ``(dowstream, upstream)`` nodes IDs of each channel segment.

        The value associated with the segment ID tuple
        ``(dowstream, upstream)`` is itself a dictionary. It has three
        key-value pairs. First, ``"ids"`` contains a list of the segment node
        IDs ordered from downstream to upstream. It includes the endpoints.
        Second, ``"distances"`` contains a list of distances upstream that
        mirrors the list in ``"ids"``. Finally, ``"color"`` is an RGBA tuple
        indicating the color for the segment.
        """
        return self._data_struct

    def _create_profile_structure(self):
        """Create the profile_IDs data structure for channel network.

        The bound attribute self._profile structure is the channel segment
        datastructure. Profile structure is a list of length
        starting_nodes. Each element of profile_structure is itself a
        list of length number of stream segments that drain to each of the
        starting nodes. Each stream segment list contains the node ids of a
        stream segment from downstream to upstream.
        """
        self._data_struct = OrderedDict()

        for i in self._starting_nodes:
            channel_segment = []
            current_node = i

            # march downstream
            while self._flow_receiver[current_node] != current_node:
                channel_segment.append(current_node)
                current_node = self._flow_receiver[current_node]
            channel_segment.append(current_node)

            channel_segment.reverse()
            segment_tuple = (current_node, i)
            self._data_struct[i] = {
                segment_tuple: {"ids": as_id_array(channel_segment)}
            }

        self._calculate_distances()
        self.assign_colors()
        self._create_flat_structures()

    def _create_flat_structures(self):
        """Create expected flattened structures for ids, distances, and colors."""
        self._nodes = []

        self._distance_along_profile = []
        self._colors = []

        for outlet_id in self._data_struct:
            seg_tuples = self._data_struct[outlet_id].keys()
            self._nodes.extend(
                [self._data_struct[outlet_id][seg]["ids"] for seg in seg_tuples]
            )
            self._distance_along_profile.extend(
                [self._data_struct[outlet_id][seg]["distances"] for seg in seg_tuples]
            )
            self._colors.extend(
                [self._data_struct[outlet_id][seg]["color"] for seg in seg_tuples]
            )

    def assign_colors(self, color_mapping=None):
        """Assign a unique color for each starting node.

        Parameters
        ----------
        color_mapping : str
            Color map name.
        """

        if color_mapping is None:
            num_watersheds = len(self._data_struct)
            norm = mpl.colors.Normalize(vmin=0, vmax=num_watersheds)
            mappable = cm.ScalarMappable(norm=norm, cmap=self._cmap)
            color_mapping = {
                outlet_id: mappable.to_rgba(idx)
                for idx, outlet_id in enumerate(self._data_struct)
            }

        for outlet_id in self._data_struct:
            for segment_tuple in self._data_struct[outlet_id]:
                self._data_struct[outlet_id][segment_tuple]["color"] = color_mapping[
                    outlet_id
                ]

    def _calculate_distances(self):
        """Get distances along the network data structure."""
        distance_upstream = calculate_flow__distance(self._grid)
        for outlet_id in self._data_struct:

            for segment_tuple in self._data_struct[outlet_id]:
                ids = self._data_struct[outlet_id][segment_tuple]["ids"]
                d = distance_upstream[ids]
                self._data_struct[outlet_id][segment_tuple]["distances"] = d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""SpeciesEvolver tests."""
from collections import Counter

import numpy as np
import pandas as pd
import pytest

from landlab import RasterModelGrid
from landlab.components import SpeciesEvolver
from landlab.components.species_evolution.base_taxon import Taxon


class TestTaxon(Taxon):
    """Taxon to test SpeciesEvolver."""

    def __init__(self, parent=None):
        super().__init__()
        self.parent = parent

    @property
    def range_mask(self):
        return np.array([False, False, False, True, True, True, False, False, False])

    def _evolve(self, dt, stage, record):
        if stage == 0:
            self._extant = False
            child_taxon = [TestTaxon(parent=self)]

        return False, child_taxon


@pytest.fixture()
def zone_example_grid():
    return RasterModelGrid((3, 3))


def test_track_taxa_and_component_attributes(zone_example_grid):
    se = SpeciesEvolver(zone_example_grid)

    # Introduce multiple taxa.
    taxa = [TestTaxon(), TestTaxon()]
    se.track_taxa(taxa)

    # Introduce a single taxon.
    taxon = TestTaxon()
    se.track_taxa(taxon)

    # Test attributes at initial time step.

    expected_df = pd.DataFrame(
        {
            "pid": 3 * [np.nan],
            "type": 3 * [TestTaxon.__name__],
            "t_first": [0, 0, 0],
            "t_final": 3 * [np.nan],
        },
        index=[0, 1, 2],
    )
    expected_df.index.name = "tid"
    expected_df["pid"] = expected_df["pid"].astype("Int64")
    expected_df["t_final"] = expected_df["t_final"].astype("Int64")
    pd.testing.assert_frame_equal(se.taxa_data_frame, expected_df, check_like=True)

    expected_df = pd.DataFrame({"time": [0], "taxa": [3]})
    pd.testing.assert_frame_equal(se.record_data_frame, expected_df, check_like=True)

    # Test attributes at a later time.

    se.run_one_step(10)

    expected_df = pd.DataFrame(
        {
            "pid": 3 * [np.nan] + [0, 1, 2],
            "type": 6 * [TestTaxon.__name__],
            "t_first": [0, 0, 0, 10, 10, 10],
            "t_final": [10, 10, 10] + 3 * [np.nan],
        },
        index=[0, 1, 2, 3, 4, 5],
    )
    expected_df.index.name = "tid"
    expected_df["pid"] = expected_df["pid"].astype("Int64")
    expected_df["t_final"] = expected_df["t_final"].astype("Int64")
    pd.testing.assert_frame_equal(se.taxa_data_frame, expected_df, check_like=True)

    expected_df = pd.DataFrame({"time": [0, 10], "taxa": [3, 3]})
    pd.testing.assert_frame_equal(se.record_data_frame, expected_df, check_like=True)


def test_get_extant_taxon_objects(zone_example_grid):
    se = SpeciesEvolver(zone_example_grid)

    introduced_taxa = [TestTaxon(), TestTaxon()]
    se.track_taxa(introduced_taxa)
    se.run_one_step(10)
    se.run_one_step(10)

    # Test no parameters.

    queried_taxa = se.get_extant_taxon_objects()
    np.testing.assert_equal(Counter(queried_taxa), Counter(se._taxon_objs))

    # Test `tids` parameter.

    queried_taxa = se.get_extant_taxon_objects(tids=[0])
    ids = [t.tid for t in queried_taxa]
    expected_ids = []
    np.testing.assert_equal(Counter(ids), Counter(expected_ids))

    queried_taxa = se.get_extant_taxon_objects(tids=[4, 5])
    ids = [t.tid for t in queried_taxa]
    expected_ids = [4, 5]
    np.testing.assert_equal(Counter(ids), Counter(expected_ids))

    # Test `time` parameter.

    queried_taxa = se.get_extant_taxon_objects(time=20)
    ids = [t.tid for t in queried_taxa]
    expected_ids = [4, 5]
    np.testing.assert_equal(Counter(ids), Counter(expected_ids))

    queried_taxa = se.get_extant_taxon_objects(time=10)
    ids = [t.tid for t in queried_taxa]
    expected_ids = []
    np.testing.assert_equal(Counter(ids), Counter(expected_ids))

    queried_taxa = se.get_extant_taxon_objects(time=30)
    ids = [t.tid for t in queried_taxa]
    expected_ids = []
    np.testing.assert_equal(Counter(ids), Counter(expected_ids))

    # Test `ancestor` parameter.

    queried_taxa = se.get_extant_taxon_objects(ancestor=1)
    ids = [t.tid for t in queried_taxa]
    expected_ids = [5]
    np.testing.assert_equal(Counter(ids), Counter(expected_ids))

    queried_taxa = se.get_extant_taxon_objects(ancestor=5)
    np.testing.assert_equal(queried_taxa, [])

    queried_taxa = se.get_extant_taxon_objects(ancestor=6)
    np.testing.assert_equal(queried_taxa, [])

    # Test multiple parameters.

    queried_taxa = se.get_extant_taxon_objects(ancestor=1, time=10)
    np.testing.assert_equal(queried_taxa, [])


def test_taxa_richness_field(zone_example_grid):
    mg = zone_example_grid

    se = SpeciesEvolver(mg)

    expected_field = np.zeros(mg.number_of_nodes)
    np.testing.assert_array_equal(mg.at_node["taxa__richness"], expected_field)

    introduced_taxa = [TestTaxon(), TestTaxon()]
    se.track_taxa(introduced_taxa)
    se.run_one_step(10)

    expected_field = np.array([0, 0, 0, 2, 2, 2, 0, 0, 0])
    np.testing.assert_array_equal(mg.at_node["taxa__richness"], expected_field)


def test_immediate_extinction(zone_example_grid):
    se = SpeciesEvolver(zone_example_grid)

    taxon = TestTaxon()
    taxon.extant = False
    se.track_taxa(taxon)

    expected_df = pd.DataFrame(
        {"pid": [np.nan], "type": [TestTaxon.__name__], "t_first": [0], "t_final": [0]},
        index=[0],
    )
    expected_df.index.name = "tid"
    expected_df["pid"] = expected_df["pid"].astype("Int64")
    expected_df["t_final"] = expected_df["t_final"].astype("Int64")
    pd.testing.assert_frame_equal(se.taxa_data_frame, expected_df, check_like=True)



# coding: utf8
# ! /usr/env/python
"""Tests for Profiler.
"""
import numpy as np
import pytest

from landlab import RasterModelGrid
from landlab.components import Profiler


def test_single_segment_profile():
    mg = RasterModelGrid((3, 5))
    core_nodes = mg.core_nodes

    endpoints = [core_nodes.min(), core_nodes.max()]
    profiler = Profiler(mg, endpoints)
    profiler.run_one_step()

    np.testing.assert_array_equal(profiler.nodes[0], core_nodes)


def test_flipped_single_segment_profile():
    mg = RasterModelGrid((3, 5))
    core_nodes = mg.core_nodes

    endpoints = [core_nodes.max(), core_nodes.min()]
    profiler = Profiler(mg, endpoints)
    profiler.run_one_step()

    np.testing.assert_array_equal(profiler.nodes[0], np.flip(core_nodes))


def test_positive_ystep_profile():
    mg = RasterModelGrid((4, 5))
    core_nodes = mg.core_nodes

    endpoints = [core_nodes.min(), core_nodes.max()]
    profiler = Profiler(mg, endpoints)
    profiler.run_one_step()

    np.testing.assert_array_equal(profiler.nodes[0], np.array([6, 7, 13]))


def test_steep_profile():
    mg = RasterModelGrid((5, 3))
    core_nodes = mg.core_nodes

    endpoints = [core_nodes.min(), core_nodes.max()]
    profiler = Profiler(mg, endpoints)
    profiler.run_one_step()
    np.testing.assert_array_equal(profiler.nodes[0], core_nodes)


def test_multi_segment_profile_structure():
    mg = RasterModelGrid((5, 5))
    core_nodes = mg.core_nodes

    endpoints = [core_nodes.min(), core_nodes.min() + 2, core_nodes.max()]
    profiler = Profiler(mg, endpoints)
    profiler.run_one_step()

    np.testing.assert_array_equal(list(profiler.data_structure.keys()), [0, 1])

    ds = profiler.data_structure

    np.testing.assert_array_equal(ds[0]["ids"], [6, 7, 8])
    np.testing.assert_array_equal(ds[1]["ids"], [8, 13, 18])

    np.testing.assert_array_equal(ds[0]["distances"], [0, 1, 2])
    np.testing.assert_array_equal(ds[1]["distances"], [2, 3, 4])


def test_endpoint_options():
    mg = RasterModelGrid((3, 5))
    core_nodes = mg.core_nodes

    endpoints = [core_nodes.min(), core_nodes.max()]
    profiler_by_nodes = Profiler(mg, endpoints)
    profiler_by_nodes.run_one_step()

    ep0 = mg.xy_of_node[core_nodes.min()]
    ep1 = mg.xy_of_node[core_nodes.max()]
    profiler_by_coords = Profiler(mg, [ep0, ep1])
    profiler_by_coords.run_one_step()

    np.testing.assert_array_equal(
        profiler_by_nodes.nodes[0], profiler_by_coords.nodes[0]
    )


def test_incorrect_endpoints_type():
    mg = RasterModelGrid((3, 5))
    core_nodes = mg.core_nodes

    endpoints = [core_nodes.min()]
    with pytest.raises(ValueError):
        Profiler(mg, endpoints)

    endpoints = [core_nodes.min(), (0, 1, 2)]
    with pytest.raises(TypeError):
        Profiler(mg, endpoints)


def test_out_of_bounds_point():
    mg = RasterModelGrid((3, 5))
    core_nodes = mg.core_nodes

    endpoints = [core_nodes.min(), 2 * core_nodes.max()]
    with pytest.raises(IndexError):
        Profiler(mg, endpoints)

    ep0 = mg.xy_of_node[core_nodes.min()]
    ep1 = 2 * mg.xy_of_node[core_nodes.max()]
    with pytest.raises(ValueError):
        Profiler(mg, [ep0, ep1])



#!/usr/bin/env python
""".. codeauthor:: Eric Hutton <huttone@colorado.edu>

.. sectionauthor:: Eric Hutton <huttone@colorado.edu>
"""

from .flexure import Flexure
from .flexure_1d import Flexure1D
from .funcs import get_flexure_parameter, subside_point_load

__all__ = ["Flexure", "Flexure1D", "get_flexure_parameter", "subside_point_load"]



import numpy as np

from ..dual import DualGraph
from ..voronoi.dual_voronoi import DualVoronoiGraph
from .radial import RadialGraph, RadialGraphLayout


class DualRadialGraph(DualGraph, RadialGraph):

    """Graph of a series of points on concentric circles.

    Examples
    --------
    >>> from landlab.graph import DualRadialGraph
    >>> graph = DualRadialGraph((1, 4), sort=True)
    >>> graph.number_of_corners
    4
    >>> graph.y_of_corner
    array([-0.5, -0.5,  0.5,  0.5])
    >>> graph.x_of_corner
    array([-0.5,  0.5, -0.5,  0.5])
    """

    def __init__(self, shape, spacing=1.0, xy_of_center=(0.0, 0.0), sort=False):
        """Create a structured grid of triangles arranged radially.

        Parameters
        ----------
        shape : tuple of int
            Shape of the graph as number of rings and number of points
            in the first ring.
        spacing : float, optional
            Spacing between rings.
        xy_of_center : tuple of float, optional
            Coordinates of the center of the grid.
        """
        try:
            spacing = float(spacing)
        except TypeError:
            raise TypeError("spacing must be a float")

        xy_of_center = tuple(np.broadcast_to(xy_of_center, 2))

        x_of_node, y_of_node = RadialGraphLayout.xy_of_node(
            shape, spacing=spacing, xy_of_center=xy_of_center
        )

        self._ring_spacing = spacing
        self._shape = tuple(shape)
        self._xy_of_center = xy_of_center

        DualVoronoiGraph.__init__(self, (y_of_node, x_of_node), sort=False)

        if sort:
            self.sort()

    @property
    def shape(self):
        return self._shape

    @property
    def spacing(self):
        return self._spacing

    @property
    def origin(self):
        return self._xy_of_center

    @property
    def xy_of_center(self):
        return self._xy_of_center



from .dump import to_netcdf
from .errors import NotRasterGridError
from .load import from_netcdf
from .read import read_netcdf
from .write import write_netcdf, write_raster_netcdf

__all__ = [
    "from_netcdf",
    "read_netcdf",
    "to_netcdf",
    "write_netcdf",
    "write_raster_netcdf",
    "NotRasterGridError",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Created on Fri Apr  8 08:32:48 2016.

@author: RCGlade
@author: dylanward
Integrated version created by D. Ward on Tue Oct 27 2020
"""

import numpy as np

from landlab import Component


class ExponentialWeathererIntegrated(Component):

    r"""
    This component implements exponential weathering of bedrock on
    hillslopes. Uses exponential soil production function in the style
    of Ahnert (1976).

    Consider that :math:`w_0` is the maximum soil production rate and
    that :math:`d^*` is the characteristic soil production depth. The
    soil production rate :math:`w` is given as a function of the soil
    depth :math:`d`,

    .. math::

        w = w_0 \exp{-\frac{d}{d^*}} \;.

    The `ExponentialWeathererIntegrated` uses the analytical solution
    for the amount of soil produced by an exponential weathering
    function over a timestep dt, and returns both the thickness of
    bedrock weathered and the thickness of soil produced. The solution
    accounts for the reduction in rate over the timestep due to the
    increasing depth. This enables accuracy over arbitrarily large
    timesteps, and better compatiblity with the `run_one_step()`
    interface.

    Compared to 'ExponentialWeatherer', upon which it is based...

    - This maintains the field I/O behavior of the original, but adds
      new return fields for the weathered thickness and soil produced
      thickness.
    - Density adjustments are needed inside the integral and the
      density ratio is intialized on instantiation. The default value
      of 1.0 assumes no change in density.
    - Returns both weathered depth of bedrock and produced depth of
      soil over the timestep.
    - The primary `soil__depth` field that is input is NOT updated by
      the component.

    This is left as an exercise for the model driver, as different
    applications may want to integrate soil depth and weathering in
    different sequences among other processes.

    - SHOULD maintain drop-in compatiblity with the plain
      :py:class:`ExponentialWeatherer <landlab.components.ExponentialWeatherer>`,
      just import and instantiate this one instead and existing code
      should work with no side effects other than the creation of the
      two additional (zeros) output fields.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import ExponentialWeathererIntegrated
    >>> mg = RasterModelGrid((5, 5))
    >>> soilz = mg.add_zeros("soil__depth", at="node")
    >>> soilrate = mg.add_ones("soil_production__rate", at="node")
    >>> expw = ExponentialWeathererIntegrated(mg)
    >>> dt = 1000
    >>> expw.run_one_step(dt)
    >>> np.allclose(mg.at_node['soil_production__rate'][mg.core_nodes], 1.)
    True
    >>> np.allclose(mg.at_node['soil_production__dt_produced_depth'][mg.core_nodes], 6.9088)
    True

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Barnhart, K., Glade, R., Shobe, C., Tucker, G. (2019). Terrainbento 1.0: a
    Python package for multi-model analysis in long-term drainage basin
    evolution. Geoscientific Model Development  12(4), 1267--1297.
    https://dx.doi.org/10.5194/gmd-12-1267-2019

    **Additional References**

    Ahnert, F. (1976). Brief description of a comprehensive three-dimensional
    process-response model of landform development Z. Geomorphol. Suppl.  25,
    29 - 49.

    Armstrong, A. (1976). A three dimensional simulation of slope forms.
    Zeitschrift für Geomorphologie  25, 20 - 28.

    """

    _name = "ExponentialWeathererIntegrated"

    _unit_agnostic = True

    _cite_as = """
    @article{barnhart2019terrain,
      author = {Barnhart, Katherine R and Glade, Rachel C and Shobe, Charles M and Tucker, Gregory E},
      title = {{Terrainbento 1.0: a Python package for multi-model analysis in long-term drainage basin evolution}},
      doi = {10.5194/gmd-12-1267-2019},
      pages = {1267---1297},
      number = {4},
      volume = {12},
      journal = {Geoscientific Model Development},
      year = {2019},
    }
    """

    _info = {
        "soil__depth": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of soil or weathered bedrock",
        },
        "soil_production__rate": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/yr",
            "mapping": "node",
            "doc": "rate of soil production at nodes",
        },
        "soil_production__dt_produced_depth": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "thickness of soil produced at nodes over time dt",
        },
        "soil_production__dt_weathered_depth": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "thickness of bedrock weathered at nodes over time dt",
        },
    }

    def __init__(
        self,
        grid,
        soil_production__maximum_rate=1.0,
        soil_production__decay_depth=1.0,
        soil_production__expansion_factor=1.0,
    ):
        """
        Parameters
        ----------
        grid: ModelGrid
            Landlab ModelGrid object
        soil_production__maximum_rate : float
            Maximum weathering rate for bare bedrock
        soil_production__decay_depth : float
            Characteristic weathering depth
        soil_production__expansion_factor : float
            Expansion ratio of regolith (from relative densities of
            rock and soil)
        """
        super().__init__(grid)

        # Store grid and parameters

        self._wstar = soil_production__decay_depth
        self._w0 = soil_production__maximum_rate
        self._fexp = soil_production__expansion_factor

        # Create fields:
        # soil depth
        self._depth = grid.at_node["soil__depth"]

        # weathering rate
        if "soil_production__rate" in grid.at_node:
            self._soil_prod_rate = grid.at_node["soil_production__rate"]
        else:
            self._soil_prod_rate = grid.add_zeros("soil_production__rate", at="node")

        # soil produced total over dt
        if "soil_production__dt_produced_depth" in grid.at_node:
            self._soil_prod_total = grid.at_node["soil_production__dt_produced_depth"]
        else:
            self._soil_prod_total = grid.add_zeros(
                "soil_production__dt_produced_depth", at="node"
            )

        # bedrock weathering total over dt
        if "soil_production__dt_weathered_depth" in grid.at_node:
            self._rock_weathered_total = grid.at_node[
                "soil_production__dt_weathered_depth"
            ]
        else:
            self._rock_weathered_total = grid.add_zeros(
                "soil_production__dt_weathered_depth", at="node"
            )

    def calc_soil_prod_rate(self):
        """Calculate soil production rate."""
        # apply exponential function
        self._soil_prod_rate[self._grid.core_nodes] = self._w0 * np.exp(
            -self._depth[self._grid.core_nodes] / self._wstar
        )

    def _calc_dt_production_total(self, dt):
        """Calculate integrated production over 1 timestep dt"""
        # analytical solution
        self._soil_prod_total[self._grid.core_nodes] = self._wstar * np.log(
            (
                self._fexp
                * self._soil_prod_rate[self._grid.core_nodes]
                * dt
                / self._wstar
            )
            + 1
        )
        # and back-convert to find rock thickness converted over the timestep:
        self._rock_weathered_total[self._grid.core_nodes] = (
            self._soil_prod_total[self._grid.core_nodes] / self._fexp
        )

    def run_one_step(self, dt=0):
        """
        Parameters
        ----------
        dt: float
            Used only for compatibility with standard run_one_step.
            If dt is not provided, the default of zero maintains backward compatibility
        """
        self.calc_soil_prod_rate()
        self._calc_dt_production_total(dt)

    @property
    def maximum_weathering_rate(self):
        """Maximum rate of weathering (m/yr)."""
        return self._w0

    @maximum_weathering_rate.setter
    def maximum_weathering_rate(self, new_val):
        if new_val <= 0:
            raise ValueError("Maximum weathering rate must be positive.")
        self._w0 = new_val



#!/usr/bin/env python

import numpy as np
import scipy.special

_POISSON = 0.25

_N_PROCS = 4


def get_flexure_parameter(h, E, n_dim, gamma_mantle=33000.0):
    """
    Calculate the flexure parameter based on some physical constants. *h* is
    the Effective elastic thickness of Earth's crust (m), *E* is Young's
    Modulus, and *n_dim* is the number of spatial dimensions for which the
    flexure parameter is used. The number of dimension must be either 1, or
    2.

    Examples
    --------
    >>> from landlab.components.flexure import get_flexure_parameter

    >>> eet = 65000.
    >>> youngs = 7e10
    >>> alpha = get_flexure_parameter(eet, youngs, 1)
    >>> print('%.3f' % round(alpha, 3))
    119965.926

    >>> alpha = get_flexure_parameter(eet, youngs, 2)
    >>> print('%.2f' % alpha)
    84828.72
    """
    D = E * pow(h, 3) / 12.0 / (1.0 - pow(_POISSON, 2))

    if n_dim not in (1, 2):
        raise ValueError("n_dim must be either 1 or 2")

    if n_dim == 2:
        alpha = pow(D / gamma_mantle, 0.25)
    else:
        alpha = pow(4.0 * D / gamma_mantle, 0.25)

    return alpha


def _calculate_distances(locs, coords):
    r = pow(coords[0][:, np.newaxis] - locs[0], 2)
    r += pow(coords[1][:, np.newaxis] - locs[1], 2)
    return np.sqrt(r, out=r)


def _calculate_deflections(load, locs, coords, alpha, out=None, gamma_mantle=33000.0):
    c = -load / (2.0 * np.pi * gamma_mantle * pow(alpha, 2.0))
    r = _calculate_distances(locs, coords) / alpha

    scipy.special.kei(r, out=r)
    np.multiply(r, c[np.newaxis, :], out=r)
    return np.sum(r, axis=1, out=out)


def subside_point_load(load, loc, coords, params=None, out=None):
    """Calculate deflection at points due a point load.

    Calculate deflections on a grid, defined by the points in the *coords*
    tuple, due to a point load of magnitude *load* applied at *loc*.

    *x* and *y* are the x and y coordinates of each node of the solution
    grid (in meters). The scalars *eet* and *youngs* define the crustal
    properties.

    Parameters
    ----------
    load : float
        Magnitude of the point load.
    loc : float or tuple
        Location of the load as either a scalar or as (*x*, *y*)
    coords : ndarray
        Array of points to calculate deflections at
    params : dict-like
        Physical parameters used for deflection calculation. Valid keys are
        - *eet*: Effective elastic thickness
        - *youngs*: Young's modulus
    out : ndarray, optional
        Array to put deflections into.

    Returns
    -------
    out : ndarray
        Array of deflections.

    Examples
    --------

    >>> from landlab.components.flexure import subside_point_load

    >>> params = dict(eet=65000., youngs=7e10)
    >>> load = 1e9

    Define a unifrom rectilinear grid.

    >>> x = np.arange(0, 10000, 100.)
    >>> y = np.arange(0, 5000, 100.)
    >>> (x, y) = np.meshgrid(x, y)
    >>> x.shape = (x.size, )
    >>> y.shape = (y.size, )

    Calculate deflections due to a load applied at position (5000., 2500.).

    >>> x = np.arange(0, 10000, 1000.)
    >>> y = np.arange(0, 5000, 1000.)
    >>> (x, y) = np.meshgrid(x, y)
    >>> x.shape = (x.size, )
    >>> y.shape = (y.size, )
    >>> dz = subside_point_load(load, (5000., 2500.), (x, y), params=params)
    >>> print('%.5g' % round(dz.sum(), 9))
    2.6267e-05
    >>> print(round(dz.min(), 9))
    5.24e-07
    >>> print(round(dz.max(), 9))
    5.26e-07

    >>> dz = subside_point_load((1e9, 1e9), ((5000., 5000.), (2500., 2500.)),
    ...                         (x, y), params=params)
    >>> print(round(dz.min(), 9) / 2.)
    5.235e-07
    >>> print(round(dz.max(), 9) / 2.)
    5.265e-07
    """
    params = params or dict(eet=6500.0, youngs=7.0e10)
    eet, youngs = params["eet"], params["youngs"]
    gamma_mantle = params.get("gamma_mantle", 33000.0)

    load = np.asarray(load).reshape((-1,))
    loc = np.asarray(loc).reshape((-1, len(load)))
    coords = np.asarray(coords)
    if coords.ndim == 1:
        coords = np.expand_dims(coords, axis=0)

    n_dim = len(loc)
    if n_dim not in (1, 2):
        raise ValueError("number of dimension must be 1 or 2")
    if len(coords) != n_dim:
        raise ValueError("number of dimensions in coordinates doesn't match loc")

    if out is None:
        out = np.empty(coords[0].size, dtype=float)

    alpha = get_flexure_parameter(eet, youngs, n_dim, gamma_mantle=gamma_mantle)

    if n_dim == 2:
        _calculate_deflections(
            load, loc, coords, alpha, out=out, gamma_mantle=gamma_mantle
        )
    else:
        x, x0 = np.meshgrid(loc[0], coords[0])
        c = load / (2.0 * alpha * gamma_mantle)
        r = abs(x - x0) / alpha
        out[:] = (c * np.exp(-r) * (np.cos(r) + np.sin(r))).sum(axis=1)

    return out



import hypothesis.extra.numpy as hynp
import numpy as np
import pytest
from hypothesis import assume, given
from hypothesis.strategies import integers
from numpy.testing import assert_array_equal

from landlab.io.shapefile.read_shapefile import _infer_data_type


@pytest.mark.parametrize("src_type", [np.int32, np.int64, float, bool])
@pytest.mark.parametrize("as_iterable", [list, tuple, np.asarray])
def test_infer_dtype(as_iterable, src_type):
    values = np.asarray([1, 2, 3], dtype=src_type)
    array = _infer_data_type(as_iterable(values))
    assert array.dtype == src_type
    assert_array_equal(array, values)


@pytest.mark.parametrize(
    "values,dst_type",
    [
        ([1, 2, 3], np.int_),
        ([1, 2, 3.0], float),
        ([1.0, 2.0, 3.0], float),
        ([1j, 2j, 3j], complex),
        ([True, True, False], bool),
    ],
)
def test_infer_dtype_from_object(values, dst_type):
    values = np.asarray(values, dtype=object)
    array = _infer_data_type(values)
    assert array.dtype == dst_type
    assert_array_equal(array, values)


@pytest.mark.parametrize(
    "values,dst_type,expected",
    [
        ([1, 2, 3.0], float, [1.0, 2.0, 3.0]),
        ([1.0, 2.0, 3j], complex, [1 + 0j, 2 + 0j, 3j]),
        ([1, 2, 3j], complex, [1 + 0j, 2 + 0j, 3j]),
        ([True, False, 1], int, [1, 0, 1]),
        ([True, False, 1.0], float, [1.0, 0.0, 1.0]),
        ([None, 1.0, 2.0], float, [np.nan, 1.0, 2.0]),
        ([None, 1.0, 2j], complex, [np.nan * 1j, 1.0, 2j]),
    ],
)
def test_infer_dtype_from_mixed(values, dst_type, expected):
    values = np.asarray(values, dtype=object)
    array = _infer_data_type(values)
    assert array.dtype == dst_type
    assert_array_equal(array, expected)


@given(
    values=hynp.arrays(
        dtype=hynp.floating_dtypes()
        | hynp.integer_dtypes()
        | hynp.complex_number_dtypes(),
        shape=hynp.array_shapes(),
        elements=integers(0, 2 ** 7 - 1),
    ),
    dst_type=hynp.floating_dtypes()
    | hynp.integer_dtypes()
    | hynp.complex_number_dtypes(),
)
@pytest.mark.parametrize("as_iterable", [list, tuple, np.asarray])
def test_dtype_keyword(as_iterable, values, dst_type):
    array = _infer_data_type(as_iterable(values), dtype=dst_type)
    assert array.dtype == dst_type
    assert_array_equal(array, values)


@given(
    values=hynp.arrays(
        dtype=hynp.floating_dtypes()
        | hynp.integer_dtypes()
        | hynp.complex_number_dtypes(),
        shape=hynp.array_shapes(),
        elements=integers(0, 2 ** 7 - 1),
    ),
    dst_type=hynp.floating_dtypes()
    | hynp.integer_dtypes()
    | hynp.complex_number_dtypes(),
)
def test_object_arrays(values, dst_type):
    assume(
        not np.issubdtype(values.dtype, np.complexfloating)
        or np.issubdtype(dst_type, np.complexfloating)
    )
    values = np.asarray(values, dtype=object)
    array = _infer_data_type(values, dtype=dst_type)
    assert array.dtype == dst_type
    assert_array_equal(array, values)


@given(
    values=hynp.arrays(
        dtype=hynp.complex_number_dtypes(),
        shape=hynp.array_shapes(),
        elements=integers(0, 8),
    ),
    dst_type=hynp.floating_dtypes() | hynp.integer_dtypes(),
)
def test_from_complex(values, dst_type):
    array = _infer_data_type(values, dtype=dst_type)
    assert_array_equal(array.real, values.real)


@given(values=hynp.arrays(dtype=hynp.array_dtypes(), shape=hynp.array_shapes()))
def test_array_not_copied(values):
    array = _infer_data_type(values, dtype=values.dtype)
    assert array is values



#! /usr/bin/env python

import os
import re

import pkg_resources
from setuptools import Extension, find_packages, setup


numpy_incl = pkg_resources.resource_filename("numpy", "core/include")


def read(filename):
    with open(filename, "r", encoding="utf-8") as fp:
        return fp.read()


long_description = u"\n\n".join(
    [
        read("README.rst"),
        read("AUTHORS.rst"),
        read("CHANGES.rst"),
    ]
)


def find_extensions(path="."):
    extensions = []
    for root, _dirs, files in os.walk(os.path.normpath(path)):
        extensions += [
            os.path.join(root, fname) for fname in files if fname.endswith(".pyx")
        ]
    return [
        Extension(re.sub(re.escape(os.path.sep), ".", ext[: -len(".pyx")]), [ext])
        for ext in extensions
    ]


setup(
    name="landlab",
    version="2.4.1",
    author="Eric Hutton",
    author_email="eric.hutton@colorado.edu",
    url="https://github.com/landlab",
    description="Plugin-based component modeling tool.",
    long_description=long_description,
    python_requires=">=3.6",
    install_requires=open("requirements.txt", "r").read().splitlines(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    keywords=[
        "bmi",
        "component modeling",
        "earth science",
        "gridding engine",
        "model coupling",
        "numerical modeling",
    ],
    packages=find_packages(),
    package_data={
        "": [
            "tests/*txt",
            "data/*asc",
            "data/*nc",
            "data/*shp",
            "test/*shx",
            "data/*dbf",
            "preciptest.in",
            "test_*/*nc",
            "test_*/*asc",
        ]
    },
    entry_points={"console_scripts": ["landlab=landlab.cmd.landlab:main"]},
    include_dirs=[numpy_incl],
    ext_modules=find_extensions("landlab"),
)



from io import BytesIO

import numpy as np
import pytest
import shapefile
from numpy.testing import assert_array_equal
from pytest import approx, raises
from shapefile import POINT, POLYLINE, ShapefileException

from landlab import ExampleData
from landlab.io.shapefile import read_shapefile


def test_read_methow(tmpdir):
    # test of the big methow network.
    shp_file = "Methow_Network.shp"
    with tmpdir.as_cwd():
        ExampleData("io/shapefile", case="methow").fetch()

        grid = read_shapefile(shp_file)
        assert grid.number_of_nodes == 721
        assert grid.number_of_links == 720

        assert "x_of_polyline" in grid.at_link
        assert "y_of_polyline" in grid.at_link

        fields = [
            "GridID",
            "Length_m",
            "ToLink",
            "usarea_km2",
            "uselev_m",
            "dselev_m",
            "Slope",
        ]
        for field in fields:
            assert field in grid.at_link

        # test for node location.
        assert grid.x_of_node[0] == approx(-1672349.0889982011)
        assert grid.y_of_node[0] == approx(1160800.240247)

        # test for node location.
        assert grid.x_of_node[25] == approx(-1677950.8375069483)
        assert grid.y_of_node[25] == approx(1173162.0319129999)

        # test for field order.
        # link 15
        assert grid.at_link["GridID"][15] == 16
        assert grid.at_link["Length_m"][15] == approx(1593.206627)
        assert grid.at_link["ToLink"][15] == 9
        assert grid.at_link["usarea_km2"][15] == 4573.3896
        assert grid.at_link["uselev_m"][15] == 305.88
        assert grid.at_link["dselev_m"][15] == 295.37
        assert grid.at_link["Slope"][15] == 0.006597

        # link 27
        assert grid.at_link["GridID"][27] == 41
        assert grid.at_link["Length_m"][27] == approx(1576.988871)
        assert grid.at_link["ToLink"][27] == 36
        assert grid.at_link["usarea_km2"][27] == 78.0597
        assert grid.at_link["uselev_m"][27] == 524.54
        assert grid.at_link["dselev_m"][27] == 399.81
        assert grid.at_link["Slope"][27] == 0.079094


def test_read_methow_subbasin(tmpdir):
    # test of the small methow network with
    shp_file = "MethowSubBasin.shp"
    points_shapefile = "MethowSubBasin_Nodes_4.shp"
    with tmpdir.as_cwd():
        ExampleData("io/shapefile", case="methow").fetch()

        grid = read_shapefile(
            shp_file, points_shapefile=points_shapefile, threshold=0.01
        )
        assert grid.number_of_nodes == 30
        assert grid.number_of_links == 29

        assert "x_of_polyline" in grid.at_link
        assert "y_of_polyline" in grid.at_link

        # verify that fields are present.
        node_fields = [
            "GridID",
            "ToLink",
            "usarea_km2",
            "uselev_m",
            "dselev_m",
            "Elev_m",
        ]
        link_fields = [
            "GridID",
            "Length_m",
            "ToLink",
            "usarea_km2",
            "uselev_m",
            "dselev_m",
            "Slope",
        ]
        for field in link_fields:
            assert field in grid.at_link
        for field in node_fields:
            assert field in grid.at_node

        # test for node location.
        assert grid.x_of_node[0] == approx(728390.38284378243)
        assert grid.y_of_node[0] == approx(5368319.8330760002)

        assert grid.x_of_node[22] == approx(725029.95616002998)
        assert grid.y_of_node[22] == approx(5374213.8330760002)

        # verify that fields are mapped correctly. choose two links and two nodes
        # to test.

        # link 1
        assert grid.at_link["GridID"][0] == 267
        assert grid.at_link["Length_m"][0] == approx(1799.167)
        assert grid.at_link["ToLink"][0] == 270
        assert grid.at_link["usarea_km2"][0] == 22.3047
        assert grid.at_link["uselev_m"][0] == 1305.7
        assert grid.at_link["dselev_m"][0] == 1232.77
        assert grid.at_link["Slope"][0] == 0.040535

        # link 16
        assert grid.at_link["GridID"][16] == 300
        assert grid.at_link["Length_m"][16] == approx(1487.342911)
        assert grid.at_link["ToLink"][16] == 263
        assert grid.at_link["usarea_km2"][16] == 75.1707
        assert grid.at_link["uselev_m"][16] == 1070.98
        assert grid.at_link["dselev_m"][16] == 986.74
        assert grid.at_link["Slope"][16] == 0.056638

        # node 1
        assert grid.at_node["GridID"][0] == 267
        assert grid.at_node["ToLink"][0] == 270.0
        assert grid.at_node["usarea_km2"][0] == 22.3047
        assert grid.at_node["uselev_m"][0] == 1305.7
        assert grid.at_node["dselev_m"][0] == 1232.77
        assert grid.at_node["Elev_m"][0] == 1304.24

        # node 29
        assert grid.at_node["GridID"][29] == 339
        assert grid.at_node["ToLink"][29] == 341.0
        assert grid.at_node["usarea_km2"][29] == 15.4314
        assert grid.at_node["uselev_m"][29] == 1534.42
        assert grid.at_node["dselev_m"][29] == 1438.68
        assert grid.at_node["Elev_m"][29] == 1535.58


def test_read_methow_subbasin_with_name_mapping_and_field_subsetting(tmpdir):
    # test of the small methow network with
    shp_file = "MethowSubBasin.shp"
    points_shapefile = "MethowSubBasin_Nodes_4.shp"

    with tmpdir.as_cwd():
        ExampleData("io/shapefile", case="methow").fetch()

        grid = read_shapefile(
            shp_file,
            points_shapefile=points_shapefile,
            node_fields=["usarea_km2", "ToLink", "Elev_m"],
            link_fields=["usarea_km2", "ToLink"],
            link_field_conversion={
                "usarea_km2": "drainage_area",
                "ToLink": "shapefile_to",
            },
            node_field_conversion={
                "usarea_km2": "drainage_area",
                "Elev_m": "topographic__elevation",
                "ToLink": "shapefile_to",
            },
            link_field_dtype={"ToLink": int},
            node_field_dtype={"ToLink": int},
            threshold=0.01,
        )

        assert grid.number_of_nodes == 30
        assert grid.number_of_links == 29

        assert "x_of_polyline" in grid.at_link
        assert "y_of_polyline" in grid.at_link

        # verify that fields are present and that some fields are not present
        node_fields = ["shapefile_to", "drainage_area", "topographic__elevation"]
        link_fields = [
            "shapefile_to",
            "drainage_area",
        ]

        not_node_fields = [
            "GridID",
            "ToLink",
            "usarea_km2",
            "uselev_m",
            "dselev_m",
            "Elev_m",
        ]
        not_link_fields = [
            "GridID",
            "Length_m",
            "ToLink",
            "usarea_km2",
            "uselev_m",
            "dselev_m",
            "Slope",
        ]

        for field in link_fields:
            assert field in grid.at_link
        for field in node_fields:
            assert field in grid.at_node
        for field in not_link_fields:
            assert field not in grid.at_link
        for field in not_node_fields:
            assert field not in grid.at_node

        # test for node location.
        assert grid.x_of_node[0] == approx(728390.38284378243)
        assert grid.y_of_node[0] == approx(5368319.8330760002)

        assert grid.x_of_node[22] == approx(725029.95616002998)
        assert grid.y_of_node[22] == approx(5374213.8330760002)

        # verify dtype changes.
        assert grid.at_link["shapefile_to"].dtype == int
        assert grid.at_node["shapefile_to"].dtype == int

        # verify that fields are mapped correctly. choose two links and two nodes
        # to test.

        # link 1
        assert grid.at_link["shapefile_to"][0] == 270
        assert grid.at_link["drainage_area"][0] == 22.3047

        # link 16
        assert grid.at_link["shapefile_to"][16] == 263
        assert grid.at_link["drainage_area"][16] == 75.1707

        # node 1
        assert grid.at_node["shapefile_to"][0] == 270
        assert grid.at_node["drainage_area"][0] == 22.3047
        assert grid.at_node["topographic__elevation"][0] == 1304.24

        # node 29
        assert grid.at_node["shapefile_to"][29] == 341
        assert grid.at_node["drainage_area"][29] == 15.4314
        assert grid.at_node["topographic__elevation"][29] == 1535.58


def test_bad_file(tmpdir):
    shp_file = "bad_file.shp"
    with raises(ShapefileException):
        read_shapefile(shp_file)


def test_points(datadir):
    shp_file = datadir / "points.shp"
    with raises(ValueError):
        read_shapefile(shp_file)


def test_multipart(datadir):
    shp_file = datadir / "multipartpolyline.shp"
    with raises(ValueError):
        read_shapefile(shp_file)


def test_bad_points():
    shp = BytesIO()
    shx = BytesIO()
    dbf = BytesIO()
    w = shapefile.Writer(shp=shp, shx=shx, dbf=dbf)
    w.shapeType = POLYLINE
    w.field("spam", "N")
    w.line([[[5, 5], [10, 10]]])
    w.record(37)
    w.line([[[5, 0], [5, 5]]])
    w.record(100)
    w.line([[[5, 5], [0, 10]]])
    w.record(239)
    w.close()

    # pass a line shapefile here insted.
    p_shp = BytesIO()
    p_shx = BytesIO()
    p_dbf = BytesIO()
    p_w = shapefile.Writer(shp=p_shp, shx=p_shx, dbf=p_dbf)
    w.shapeType = POLYLINE
    p_w.field("spam", "N")
    p_w.line([[[5, 5], [10, 10]]])
    p_w.record(37)
    p_w.line([[[5, 0], [5, 5]]])
    p_w.record(100)
    p_w.line([[[5, 5], [0, 10]]])
    p_w.record(239)
    p_w.close()

    with raises(ValueError):
        read_shapefile(shp, dbf=dbf, points_shapefile=p_shp, points_dbf=p_dbf)


def test_points_but_too_far():
    shp = BytesIO()
    shx = BytesIO()
    dbf = BytesIO()
    w = shapefile.Writer(shp=shp, shx=shx, dbf=dbf)
    w.shapeType = POLYLINE
    w.field("spam", "N")
    w.line([[[5, 5], [10, 10]]])
    w.record(37)
    w.line([[[5, 0], [5, 5]]])
    w.record(100)
    w.line([[[5, 5], [0, 10]]])
    w.record(239)
    w.close()

    # make a
    p_shp = BytesIO()
    p_shx = BytesIO()
    p_dbf = BytesIO()
    p_w = shapefile.Writer(shp=p_shp, shx=p_shx, dbf=p_dbf)
    p_w.shapeType = POINT
    p_w.field("eggs", "N")
    p_w.point(5, 0)
    p_w.record(2)
    p_w.point(5, 5)
    p_w.record(4)
    p_w.point(0, 10)
    p_w.record(8)
    p_w.point(12, 10)
    p_w.record(6)
    p_w.close()

    with raises(ValueError):
        read_shapefile(shp, dbf=dbf, points_shapefile=p_shp, points_dbf=p_dbf)


def test_points_but_not_one_one():
    shp = BytesIO()
    shx = BytesIO()
    dbf = BytesIO()
    w = shapefile.Writer(shp=shp, shx=shx, dbf=dbf)
    w.shapeType = POLYLINE
    w.field("spam", "N")
    w.line([[[5, 5], [10, 10]]])
    w.record(37)
    w.line([[[5, 0], [5, 5]]])
    w.record(100)
    w.line([[[5, 5], [0, 10]]])
    w.record(239)
    w.close()

    # make a
    p_shp = BytesIO()
    p_shx = BytesIO()
    p_dbf = BytesIO()
    p_w = shapefile.Writer(shp=p_shp, shx=p_shx, dbf=p_dbf)
    p_w.shapeType = POINT
    p_w.field("eggs", "N")
    p_w.point(5, 0)
    p_w.record(2)
    p_w.point(5, 5)
    p_w.record(4)
    p_w.point(0, 10)
    p_w.record(8)
    p_w.point(10, 10)
    p_w.record(6)
    p_w.point(10, 10)
    p_w.record(7)
    p_w.close()

    with raises(ValueError):
        read_shapefile(shp, dbf=dbf, points_shapefile=p_shp, points_dbf=p_dbf)


def test_points_but_one_missing():
    shp = BytesIO()
    shx = BytesIO()
    dbf = BytesIO()
    w = shapefile.Writer(shp=shp, shx=shx, dbf=dbf)
    w.shapeType = POLYLINE
    w.field("spam", "N")
    w.line([[[5, 5], [10, 10]]])
    w.record(37)
    w.line([[[5, 0], [5, 5]]])
    w.record(100)
    w.line([[[5, 5], [0, 10]]])
    w.record(239)
    w.close()

    # make a
    p_shp = BytesIO()
    p_shx = BytesIO()
    p_dbf = BytesIO()
    p_w = shapefile.Writer(shp=p_shp, shx=p_shx, dbf=p_dbf)
    p_w.shapeType = POINT
    p_w.field("eggs", "N")
    p_w.point(5, 0)
    p_w.record(2)
    p_w.point(5, 5)
    p_w.record(4)
    p_w.point(0, 10)
    p_w.record(8)
    p_w.close()

    with raises(ValueError):
        read_shapefile(shp, dbf=dbf, points_shapefile=p_shp, points_dbf=p_dbf)


def test_simple_reorder():

    orders = [(0, 1, 2), (0, 2, 1), (1, 0, 2), (1, 2, 0), (2, 0, 1), (2, 1, 0)]

    lines = [
        [[[5, 5], [10, 10]]],
        [[[5, 0], [5, 5]]],
        [[[5, 5], [0, 10]]],
    ]

    records = [37, 100, 239]

    point_orders = [
        (0, 1, 2, 3),
        (0, 2, 3, 1),
        (0, 3, 2, 1),
        (0, 2, 3, 1),
        (1, 0, 2, 3),
        (1, 0, 3, 2),
        (1, 2, 3, 0),
        (1, 3, 2, 0),
    ]

    points = [(5, 0), (5, 5), (0, 10), (10, 10)]
    point_records = [2, 4, 8, 6]

    for order in orders:

        for p_order in point_orders:

            shp = BytesIO()
            shx = BytesIO()
            dbf = BytesIO()

            w = shapefile.Writer(shp=shp, shx=shx, dbf=dbf)

            w.shapeType = POLYLINE
            w.field("spam", "N")

            for o in order:
                w.line(lines[o])
                w.record(records[o])
            w.close()

            p_shp = BytesIO()
            p_shx = BytesIO()
            p_dbf = BytesIO()
            p_w = shapefile.Writer(shp=p_shp, shx=p_shx, dbf=p_dbf)
            p_w.shapeType = POINT
            p_w.field("eggs", "N")
            for po in p_order:
                p_w.point(*points[po])
                p_w.record(point_records[po])
            p_w.close()

            grid = read_shapefile(
                shp, dbf=dbf, points_shapefile=p_shp, points_dbf=p_dbf
            )

            assert_array_equal(grid.nodes, np.array([0, 1, 2, 3]))
            assert_array_equal(grid.x_of_node, np.array([5.0, 5.0, 0.0, 10.0]))
            assert_array_equal(grid.y_of_node, np.array([0.0, 5.0, 10.0, 10.0]))
            assert_array_equal(grid.nodes_at_link, np.array([[0, 1], [2, 1], [1, 3]]))
            assert "spam" in grid.at_link
            assert_array_equal(grid.at_link["spam"], np.array([100, 239, 37]))

            del grid, w, shp, shx, dbf


@pytest.mark.parametrize(
    "dtype", [int, float, bool, np.int32, np.int64, np.complex128, object, None]
)
def test_field_dtype(tmpdir, dtype):
    # test of the small methow network with
    shp_file = "MethowSubBasin.shp"
    points_shapefile = "MethowSubBasin_Nodes_4.shp"

    with tmpdir.as_cwd():
        ExampleData("io/shapefile", case="methow").fetch()

        grid = read_shapefile(
            shp_file,
            points_shapefile=points_shapefile,
            node_fields=["usarea_km2", "ToLink", "Elev_m"],
            link_fields=["usarea_km2", "ToLink"],
            link_field_conversion={
                "usarea_km2": "drainage_area",
                "ToLink": "shapefile_to",
            },
            node_field_conversion={
                "usarea_km2": "drainage_area",
                "Elev_m": "topographic__elevation",
                "ToLink": "shapefile_to",
            },
            link_field_dtype={"ToLink": dtype},
            node_field_dtype={"ToLink": dtype},
            threshold=0.01,
        )

    if dtype is None:
        dtype = float

    assert grid.at_node["shapefile_to"].dtype == dtype
    assert grid.at_link["shapefile_to"].dtype == dtype



#! /usr/bin/env python
"""Calculate slope aspects on a `RasterModelGrid`."""


def _one_line_slopes(input_array, grid, vals):
    node = input_array[0]
    diagonals = input_array[5:]
    neighbors = input_array[1:5]

    if not grid.status_at_node[node] == 0:
        raise IndexError("One or more of the provided nodes was closed!")

    try:
        slope_we = (
            (vals[diagonals[1]] + 2.0 * vals[neighbors[2]] + vals[diagonals[2]])
            - (vals[diagonals[0]] + 2.0 * vals[neighbors[0]] + vals[diagonals[3]])
        ) / (8.0 * grid.dx)
        slope_sn = (
            (vals[diagonals[2]] + 2.0 * vals[neighbors[3]] + vals[diagonals[3]])
            - (vals[diagonals[1]] + 2.0 * vals[neighbors[:, 1]] + vals[diagonals[0]])
        ) / (8.0 * grid.dy)
        return slope_we, slope_sn
    except IndexError:
        C = vals[node]
        weighting_verticals = 4.0
        weighting_horizontals = 4.0
        try:
            vertical_grad = (vals[neighbors[3]] - vals[neighbors[1]]) / (2.0 * grid.dy)
        except IndexError:
            try:
                vertical_grad = (C - vals[neighbors[1]]) / grid.dy
            except IndexError:
                try:
                    vertical_grad = (vals[neighbors[3]] - C) / grid.dy
                except IndexError:
                    vertical_grad = 0.0
                    weighting_verticals -= 2.0
        try:
            horizontal_grad = (vals[neighbors[2]] - vals[neighbors[0]]) / (
                2.0 * grid.dx
            )
        except IndexError:
            try:
                horizontal_grad = (C - vals[neighbors[0]]) / grid.dx
            except IndexError:
                try:
                    horizontal_grad = (vals[neighbors[2]] - C) / grid.dx
                except IndexError:
                    horizontal_grad = 0.0
                    weighting_horizontals -= 2.0
        try:
            left_grad = (vals[diagonals[2]] - vals[diagonals[1]]) / (2.0 * grid.dx)
        except IndexError:
            try:
                C = vals[neighbors[2]]
            except IndexError:
                left_grad = 0.0
                weighting_verticals -= 1.0
            else:
                try:
                    left_grad = (C - vals[diagonals[1]]) / grid.dx
                except IndexError:
                    left_grad = (vals[diagonals[2]] - C) / grid.dx
        try:
            right_grad = (vals[diagonals[3]] - vals[diagonals[0]]) / (2.0 * grid.dx)
        except IndexError:
            try:
                C = vals[neighbors[0]]
            except IndexError:
                right_grad = 0.0
                weighting_verticals -= 1.0
            else:
                try:
                    right_grad = (C - vals[diagonals[0]]) / grid.dx
                except IndexError:
                    right_grad = (vals[diagonals[3]] - C) / grid.dx
        try:
            top_grad = (vals[diagonals[1]] - vals[diagonals[0]]) / (2.0 * grid.dy)
        except IndexError:
            try:
                C = vals[neighbors[1]]
            except IndexError:
                top_grad = 0.0
                weighting_horizontals -= 1.0
            else:
                try:
                    top_grad = (C - vals[diagonals[0]]) / grid.dy
                except IndexError:
                    top_grad = (vals[diagonals[1]] - C) / grid.dy
        try:
            bottom_grad = (vals[diagonals[2]] - vals[diagonals[3]]) / (2.0 * grid.dy)
        except IndexError:
            try:
                C = vals[neighbors[3]]
            except IndexError:
                bottom_grad = 0.0
                weighting_horizontals -= 1.0
            else:
                try:
                    bottom_grad = (C - vals[diagonals[3]]) / grid.dy
                except IndexError:
                    bottom_grad = (vals[diagonals[2]] - C) / grid.dy

        slope_we = (
            top_grad + 2.0 * horizontal_grad + bottom_grad
        ) / weighting_horizontals
        slope_sn = (left_grad + 2.0 * vertical_grad + right_grad) / weighting_verticals

        return slope_we, slope_sn



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.data_record.data_record.DataRecord
Dimension = time

Last updated 10/18/2018

"""

from landlab import RasterModelGrid

grid = RasterModelGrid((3, 3))


def test_dr_nodim_name(dr_nodim):
    assert dr_nodim._name == "DataRecord"



"""Define a graph of nodes-links-patches and its dual.

This class should not be used directly. Instead, it should be used as a
base class when defining other types of graphs.
"""
import inspect
from functools import lru_cache

import numpy as np

from ..core.utils import as_id_array
from .graph import Graph
from .graph_convention import ConventionConverter
from .sort.sort import reverse_one_to_one


class DualGraphMeta(type):
    def __init__(cls, name, bases, dct):
        type.__init__(cls, name, bases, dct)

        converter = ConventionConverter("cfc")
        for name, prop in inspect.getmembers(cls, inspect.isdatadescriptor):
            new_name = converter.conform(name, "nlp")
            if hasattr(cls, new_name):
                continue

            fdoc = inspect.getdoc(prop)
            if fdoc:
                fdoc = inspect.cleandoc(
                    """{0}

                    See Also
                    --------
                    Graph.{1}
                    """.format(
                        converter.conform(fdoc.splitlines()[0], "nlp"), name
                    )
                )

            setattr(
                cls,
                new_name,
                property(lambda x, name=name: getattr(x._dual, name), None, None, fdoc),
            )


class DualGraph(metaclass=DualGraphMeta):
    @property
    def dual(self):
        return self._dual

    @property
    def node_at_cell(self):
        return self.ds["node_at_cell"].values

    @property
    def nodes_at_face(self):
        return self.ds["nodes_at_face"].values

    @property
    @lru_cache()
    def cell_at_node(self):
        return reverse_one_to_one(self.node_at_cell, minlength=self.number_of_nodes)

    @property
    @lru_cache()
    def link_at_face(self):
        return self._create_link_at_face()

    def _create_link_at_face(self):

        link_at_nodes = {}
        for link, pair in enumerate(self.nodes_at_link):
            # pair.sort()
            link_at_nodes[tuple(np.sort(pair))] = link

        link_at_face = np.full((self.number_of_faces,), -1, dtype=int)
        # for face, pair in enumerate(self._nodes_at_face):
        for face, pair in enumerate(self.nodes_at_face):
            # pair.sort()
            link_at_face[face] = link_at_nodes[tuple(np.sort(pair))]
        self._link_at_face = link_at_face

        return self._link_at_face

    @property
    @lru_cache()
    def face_at_link(self):
        return reverse_one_to_one(self.link_at_face, minlength=self.number_of_links)

    def sort(self):
        from .sort.ext.remap_element import remap_graph_element

        sorted_nodes, sorted_links, sorted_patches = Graph.sort(self)
        sorted_corners, sorted_faces, sorted_cells = self.dual.sort()

        with self.thawed():
            self.node_at_cell[:] = self.node_at_cell[sorted_cells]
            self.nodes_at_face[:] = self.nodes_at_face[sorted_faces]

            remap_graph_element(
                as_id_array(self.node_at_cell), as_id_array(np.argsort(sorted_nodes))
            )
            remap_graph_element(
                as_id_array(self.nodes_at_face).reshape((-1,)),
                as_id_array(np.argsort(sorted_nodes)),
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Created on Fri Apr  8 08:32:48 2016.

@author: RCGlade
"""

import numpy as np

from landlab import Component


class ExponentialWeatherer(Component):

    r"""
    This component implements exponential weathering of bedrock on hillslopes.
    Uses exponential soil production function in the style of Ahnert (1976).

    Consider that :math:`w_0` is the maximum soil production rate and
    that :math:`d^*` is the characteristic soil production depth. The
    soil production rate :math:`w` is given as a function of the soil
    depth :math:`d`,

    .. math::

        w = w_0 \exp{-\frac{d}{d^*}} \;.

    The `ExponentialWeatherer` only calculates soil production at core nodes.

    An alternative version which uses the analytical integral of
    production through time is available at the component
    :py:class:`ExponentialWeathererIntegrated <landlab.components.ExponentialWeathererIntegrated>`.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import ExponentialWeatherer
    >>> mg = RasterModelGrid((5, 5))
    >>> soilz = mg.add_zeros("soil__depth", at="node")
    >>> soilrate = mg.add_ones("soil_production__rate", at="node")
    >>> expw = ExponentialWeatherer(mg)
    >>> expw.calc_soil_prod_rate()
    >>> np.allclose(mg.at_node['soil_production__rate'], 1.)
    True

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Barnhart, K., Glade, R., Shobe, C., Tucker, G. (2019). Terrainbento 1.0: a
    Python package for multi-model analysis in long-term drainage basin
    evolution. Geoscientific Model Development  12(4), 1267--1297.
    https://dx.doi.org/10.5194/gmd-12-1267-2019

    **Additional References**

    Ahnert, F. (1976). Brief description of a comprehensive three-dimensional
    process-response model of landform development Z. Geomorphol. Suppl.  25,
    29 - 49.

    Armstrong, A. (1976). A three dimensional simulation of slope forms.
    Zeitschrift für Geomorphologie  25, 20 - 28.

    """

    _name = "ExponentialWeatherer"

    _unit_agnostic = True

    _cite_as = """
    @article{barnhart2019terrain,
      author = {Barnhart, Katherine R and Glade, Rachel C and Shobe, Charles M and Tucker, Gregory E},
      title = {{Terrainbento 1.0: a Python package for multi-model analysis in long-term drainage basin evolution}},
      doi = {10.5194/gmd-12-1267-2019},
      pages = {1267---1297},
      number = {4},
      volume = {12},
      journal = {Geoscientific Model Development},
      year = {2019},
    }
    """

    _info = {
        "soil__depth": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of soil or weathered bedrock",
        },
        "soil_production__rate": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/yr",
            "mapping": "node",
            "doc": "rate of soil production at nodes",
        },
    }

    def __init__(
        self, grid, soil_production__maximum_rate=1.0, soil_production__decay_depth=1.0
    ):
        """
        Parameters
        ----------
        grid: ModelGrid
            Landlab ModelGrid object
        soil_production__maximum_rate : float
            Maximum weathering rate for bare bedrock
        soil_production__decay_depth : float
            Characteristic weathering depth
        """
        super().__init__(grid)

        # Store grid and parameters

        self._wstar = soil_production__decay_depth
        self._w0 = soil_production__maximum_rate

        # Create fields:
        # soil depth
        self._depth = grid.at_node["soil__depth"]

        # weathering rate
        if "soil_production__rate" in grid.at_node:
            self._soil_prod_rate = grid.at_node["soil_production__rate"]
        else:
            self._soil_prod_rate = grid.add_zeros("soil_production__rate", at="node")

    def calc_soil_prod_rate(self):
        """Calculate soil production rate."""
        # apply exponential function
        self._soil_prod_rate[self._grid.core_nodes] = self._w0 * np.exp(
            -self._depth[self._grid.core_nodes] / self._wstar
        )

    def run_one_step(self):
        """

        Parameters
        ----------
        dt: float
            Used only for compatibility with standard run_one_step.
        """
        self.calc_soil_prod_rate()

    @property
    def maximum_weathering_rate(self):
        """Maximum rate of weathering (m/yr)."""
        return self._w0

    @maximum_weathering_rate.setter
    def maximum_weathering_rate(self, new_val):
        if new_val <= 0:
            raise ValueError("Maximum weathering rate must be positive.")
        self._w0 = new_val



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 17:12:24 2021

@author: gtucker
"""

from numpy.testing import assert_array_almost_equal, assert_array_equal

from landlab import RasterModelGrid
from landlab.components import SimpleSubmarineDiffuser


def test_diffusivity_vs_depth():
    """
    Make sure the diffusivity-vs-depth function is correct.

    This test checks that the diffusivity should equal the shallow water
    diffusivity at the shoreline (when tidal range is zero), and at the wave
    base depth (here 50 m). It should be about half the nominal value at a
    depth of 84.657359 m (1 - ln(1/2) times the wave base depth). It should be
    near-zero for negative depths.

    When the tidal range is nonzero, the value at a depth of zero (shoreline)
    should be half the shallow-water diffusivity.
    """
    grid = RasterModelGrid((3, 3), xy_spacing=40.0)
    topo = grid.add_zeros("topographic__elevation", at="node")
    topo[3] = -84.657359
    topo[4] = -50.0
    topo[5] = 50.0

    ssd = SimpleSubmarineDiffuser(
        grid, tidal_range=0.0, wave_base=50.0, shallow_water_diffusivity=100.0
    )
    ssd.calc_diffusion_coef()
    assert_array_almost_equal(grid.at_node["kd"][2:6], [100.0, 50.0, 100.0, 1.0e-20])

    ssd = SimpleSubmarineDiffuser(
        grid, tidal_range=2.0, wave_base=50.0, shallow_water_diffusivity=100.0
    )
    ssd.calc_diffusion_coef()
    assert_array_almost_equal(grid.at_node["kd"][2:6], [50.0, 50.0, 100.0, 1.0e-20])


def test_one_step_shallow():
    r"""
    Test one time step of diffusion in shallow water.

    The test uses a submarine topography above wave base, with a cross-section
    that looks like this:

        o_./'\._o

    where o is an open-boundary node, . and ' indicate a core node, and the
    slope of the / and \ links is 0.1. The other boundary nodes are closed.
    With a diffusivity coefficient of 100 m2/y, the flux along the sloping
    links should be 10 m2/y. Let the node spacing be 40 m. The height of most
    of the nodes will be -10 m, and the central node -6 m. With a time step of
    2 years, the central node should lose 20 m2 to each side, for a total of
    40 m2. This translates into an elevation reduction of 1 m. The nodes on
    either side each gain 40 m2, for an elevation gain of 0.5 m. Therefore the
    3 core nodes should change from elevatios of [-10, -6, -10] to
    [-9.5, -7, -9.5].
    """
    grid = RasterModelGrid((3, 5), xy_spacing=40.0)
    grid.set_closed_boundaries_at_grid_edges(
        right_is_closed=False,
        top_is_closed=True,
        left_is_closed=False,
        bottom_is_closed=True,
    )
    topo = grid.add_zeros("topographic__elevation", at="node")
    topo[:] = -10.0
    topo[7] = -6.0

    ssd = SimpleSubmarineDiffuser(
        grid, tidal_range=0.0, wave_base=100.0, shallow_water_diffusivity=100.0
    )
    ssd.run_one_step(dt=2.0)
    assert_array_equal(topo[6:9], [-9.5, -7, -9.5])
    assert_array_equal(
        grid.at_node["sediment_deposit__thickness"][6:9], [0.5, -1.0, 0.5]
    )


def test_one_step_deep():
    """
    Test one time step of diffusion in deep water (below wave base).

    This test is similar to the shallow-water test but operates below the wave
    base. Let the wave base be 50 m. Using the exponential formulation for
    depth-decay of transport, the transport coefficient at depth D is
    K0 exp( -(D - Dwb) / Dwb ), where Dwb is the wave-base depth, and K0 is the
    shallow-water transport coefficient (diffusivity). If D = 85 m, then the
    transport coefficient is about 0.5 K0.
    """
    grid = RasterModelGrid((3, 5), xy_spacing=40.0)
    grid.set_closed_boundaries_at_grid_edges(
        right_is_closed=False,
        top_is_closed=True,
        left_is_closed=False,
        bottom_is_closed=True,
    )
    topo = grid.add_zeros("topographic__elevation", at="node")
    topo[:] = -87.0
    topo[7] = -83.0

    ssd = SimpleSubmarineDiffuser(
        grid, tidal_range=0.0, wave_base=50.0, shallow_water_diffusivity=100.0
    )
    ssd.run_one_step(dt=2.0)
    assert_array_almost_equal(topo[6:9], [-86.741574, -83.516851, -86.741574])


def test_depth_function():
    """
    Test the depth-weighting function.

    If tidal range is zero, the weight should be 1 where water depth is >=0,
    and 0 otherwise. If tidal range is >0, the weighting function should be 0.5
    where depth = 0, about 0.8808 where depth equals tidal range, and about
    0.1192 where depth equals minus one tidal range.
    """
    grid = RasterModelGrid((3, 4))
    topo = grid.add_zeros("topographic__elevation", at="node")
    topo[5] = -2.0
    topo[6] = 2.0

    ssd = SimpleSubmarineDiffuser(grid, tidal_range=0.0)
    depth = -topo
    df = ssd.depth_function(depth)
    assert_array_equal(df, [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0])

    ssd = SimpleSubmarineDiffuser(grid, tidal_range=2.0)
    df = ssd.depth_function(depth)
    assert_array_almost_equal(
        df, [0.5, 0.5, 0.5, 0.5, 0.5, 0.880797, 0.119203, 0.5, 0.5, 0.5, 0.5, 0.5]
    )



#! /usr/bin/env python
"""Decorators for TheLandlab package."""
import re


def camel_case(text, sep=None):
    """Convert to camel case.

    Convert *text* to camel case. Use the *sep* keyword to specify the word
    separator. The default is to split on whitespace.

    >>> from landlab.framework.decorators import camel_case
    >>> camel_case("eric idle")
    'EricIdle'
    >>> camel_case("terry_gilliam", sep="_")
    'TerryGilliam'
    >>> camel_case("MONTY Python")
    'MONTYPython'
    >>> camel_case("GrahamChapman")
    'GrahamChapman'
    """
    return "".join([word[0].upper() + word[1:] for word in text.split(sep)])


def snake_case(text):
    """Convert camel case to snake case.

    Examples
    --------
    >>> from landlab.framework.decorators import snake_case
    >>> snake_case("EricIdle")
    'eric_idle'
    >>> snake_case("MONTYPython")
    'monty_python'
    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()



from .base import ModelGrid
from .create import create_grid
from .hex import HexModelGrid
from .network import NetworkModelGrid
from .radial import RadialModelGrid
from .raster import RasterModelGrid
from .voronoi import VoronoiDelaunayGrid

__all__ = [
    "ModelGrid",
    "HexModelGrid",
    "RadialModelGrid",
    "RasterModelGrid",
    "VoronoiDelaunayGrid",
    "NetworkModelGrid",
    "create_grid",
]



#! /usr/bin/env python
from enum import IntEnum, unique

import numpy as np

from .nodestatus import NodeStatus


@unique
class LinkStatus(IntEnum):
    """Define the link types"""

    #: Indicate a link is *active*, and can carry flux
    ACTIVE = 0
    #: Indicate a link has a fixed (gradient) value, & behaves as a boundary
    FIXED = 2
    #: Indicate a link is *inactive*, and cannot carry flux
    INACTIVE = 4


def is_fixed_link(node_status_at_link):
    """Find links that are fixed.

    A link is fixed if it connects a core node with a fixed value
    boundary node.

    Parameters
    ----------
    node_status_at_link : ndarray of int, shape `(n_links, 2)`
        Node status a link tail and head.

    Returns
    -------
    ndarray of bool, shape `(n_links, )`
        True if link is fixed.

    Examples
    --------
    >>> from landlab.grid.linkstatus import is_fixed_link
    >>> from landlab import NodeStatus
    >>> is_fixed_link([NodeStatus.CORE, NodeStatus.FIXED_GRADIENT])
    array([ True], dtype=bool)

    >>> is_fixed_link([NodeStatus.CORE, NodeStatus.FIXED_VALUE])
    array([False], dtype=bool)

    >>> is_fixed_link(
    ...     [
    ...         [NodeStatus.FIXED_GRADIENT, NodeStatus.CORE],
    ...         [NodeStatus.CORE, NodeStatus.CORE]
    ...     ]
    ... )
    array([ True, False], dtype=bool)
    """
    node_status_at_link = np.asarray(node_status_at_link).reshape((-1, 2))

    is_core_node = node_status_at_link == NodeStatus.CORE
    is_fixed_gradient_node = node_status_at_link == NodeStatus.FIXED_GRADIENT

    return (is_core_node[:, 0] & is_fixed_gradient_node[:, 1]) | (
        is_fixed_gradient_node[:, 0] & is_core_node[:, 1]
    )


def is_inactive_link(node_status_at_link):
    """Find links that are inactive.

    A link is inactive if it connects two boundary nodes or one of
    its nodes is closed.

    Parameters
    ----------
    node_status_at_link : ndarray of int, shape `(n_links, 2)`
        Node status a link tail and head.

    Returns
    -------
    ndarray of bool, shape `(n_links, )`
        True if link is isactive.

    Examples
    --------
    >>> from landlab.grid.linkstatus import is_inactive_link
    >>> from landlab import NodeStatus
    >>> is_inactive_link([NodeStatus.CORE, NodeStatus.CLOSED])
    array([ True], dtype=bool)

    >>> is_inactive_link([NodeStatus.FIXED_GRADIENT, NodeStatus.FIXED_VALUE])
    array([ True], dtype=bool)

    >>> is_inactive_link([[NodeStatus.FIXED_GRADIENT, NodeStatus.CLOSED],
    ...                   [NodeStatus.CORE, NodeStatus.CORE]])
    array([ True, False], dtype=bool)
    """
    node_status_at_link = np.asarray(node_status_at_link).reshape((-1, 2))

    is_core = node_status_at_link == NodeStatus.CORE
    is_fixed_value = node_status_at_link == NodeStatus.FIXED_VALUE
    is_fixed_gradient = node_status_at_link == NodeStatus.FIXED_GRADIENT
    is_closed = node_status_at_link == NodeStatus.CLOSED
    is_boundary_node = is_fixed_value | is_fixed_gradient | is_closed

    return (
        (is_boundary_node[:, 0] & is_boundary_node[:, 1])
        | (is_closed[:, 0] & is_core[:, 1])
        | (is_core[:, 0] & is_closed[:, 1])
    )


def is_active_link(node_status_at_link):
    """Find links that are active.

    A link is active if it connects a core node with another core
    node or a fixed value boundary.

    Parameters
    ----------
    node_status_at_link : ndarray of int, shape `(n_links, 2)`
        Node status a link tail and head.

    Returns
    -------
    ndarray of bool, shape `(n_links, )`
        True if link is isactive.

    Examples
    --------
    >>> from landlab.grid.linkstatus import is_active_link
    >>> from landlab import NodeStatus
    >>> is_active_link([NodeStatus.CORE, NodeStatus.FIXED_GRADIENT])
    array([False], dtype=bool)

    >>> is_active_link([NodeStatus.CORE, NodeStatus.FIXED_VALUE])
    array([ True], dtype=bool)

    >>> is_active_link(
    ...     [
    ...         [NodeStatus.FIXED_GRADIENT, NodeStatus.CORE],
    ...         [NodeStatus.CORE, NodeStatus.CORE]
    ...     ]
    ... )
    array([False, True], dtype=bool)
    """
    node_status_at_link = np.asarray(node_status_at_link).reshape((-1, 2))

    is_core_node = node_status_at_link == NodeStatus.CORE
    is_fixed_value_node = node_status_at_link == NodeStatus.FIXED_VALUE
    return (
        (is_core_node[:, 0] & is_core_node[:, 1])
        | (is_core_node[:, 0] & is_fixed_value_node[:, 1])
        | (is_fixed_value_node[:, 0] & is_core_node[:, 1])
    )


def set_status_at_link(node_status_at_link, out=None):
    n_links = len(node_status_at_link)

    if out is None:
        out = np.full(n_links, 255, dtype=np.uint8)

    _is_fixed_link = is_fixed_link(node_status_at_link)
    _is_active_link = is_active_link(node_status_at_link)
    _is_inactive_link = is_inactive_link(node_status_at_link)

    assert np.all(
        np.sum(np.vstack((_is_active_link, _is_inactive_link, _is_fixed_link)), axis=0)
        == 1
    )

    out[_is_inactive_link] = LinkStatus.INACTIVE
    out[_is_active_link] = LinkStatus.ACTIVE
    out[_is_fixed_link] = LinkStatus.FIXED

    return out



import numpy as np
cimport numpy as np
cimport cython


DTYPE_INT = int
ctypedef np.int_t DTYPE_INT_t

DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t


@cython.boundscheck(False)
cpdef _add_to_stack(DTYPE_INT_t l, DTYPE_INT_t j,
                    np.ndarray[DTYPE_INT_t, ndim=1] s,
                    np.ndarray[DTYPE_INT_t, ndim=1] delta,
                    np.ndarray[DTYPE_INT_t, ndim=1] donors):

    """
    Adds node l to the stack and increments the current index (j).
    """
    cdef int m, n, delta_l, delta_lplus1

    s[j] = l
    j += 1
    delta_l = delta[l]
    delta_lplus1 = delta[l+1]

    for n in range(delta_l, delta_lplus1):
        m = donors[n]
        if m != l:
            j = _add_to_stack(m, j, s, delta, donors)

    return j


@cython.boundscheck(False)
cpdef _accumulate_to_n(DTYPE_INT_t np, DTYPE_INT_t q,
                       np.ndarray[DTYPE_INT_t, ndim=1] s,
                       np.ndarray[DTYPE_INT_t, ndim=2] r,
                       np.ndarray[DTYPE_FLOAT_t, ndim=2] p,
                       np.ndarray[DTYPE_FLOAT_t, ndim=1] drainage_area,
                       np.ndarray[DTYPE_FLOAT_t, ndim=1] discharge):
    """
    Accumulates drainage area and discharge, permitting transmission losses.
    """
    cdef int donor, recvr, i, v
    cdef float accum, proportion

    # Iterate backward through the list, which means we work from upstream to
    # downstream.
    for i in range(np-1, -1, -1):
        donor = s[i]
        for v in range(q):
            recvr = r[donor, v]
            proportion = p[donor, v]
            if proportion > 0.:
                if donor != recvr:
                    drainage_area[recvr] += proportion*drainage_area[donor]
                    accum = discharge[recvr] + proportion*discharge[donor]
                    if accum < 0.:
                        accum = 0.
                    discharge[recvr] = accum


@cython.boundscheck(False)
cpdef _accumulate_bw(DTYPE_INT_t np,
                     np.ndarray[DTYPE_INT_t, ndim=1] s,
                     np.ndarray[DTYPE_INT_t, ndim=1] r,
                     np.ndarray[DTYPE_FLOAT_t, ndim=1] drainage_area,
                     np.ndarray[DTYPE_FLOAT_t, ndim=1] discharge):
    """
    Accumulates drainage area and discharge, permitting transmission losses.
    """
    cdef int donor, recvr, i
    cdef float accum

    # Iterate backward through the list, which means we work from upstream to
    # downstream.
    for i in range(np-1, -1, -1):
        donor = s[i]
        recvr = r[donor]
        if donor != recvr:
            drainage_area[recvr] += drainage_area[donor]
            accum = discharge[recvr] + discharge[donor]
            if accum < 0.:
                accum = 0.
            discharge[recvr] = accum


@cython.boundscheck(False)
cpdef _make_donors(DTYPE_INT_t np,
                   np.ndarray[DTYPE_INT_t, ndim=1] w,
                   np.ndarray[DTYPE_INT_t, ndim=1] D,
                   np.ndarray[DTYPE_INT_t, ndim=1] delta,
                   np.ndarray[DTYPE_INT_t, ndim=1] r):
    """Determines number of donors"""
    cdef int ri, i
    for i in range(np):
        ri = r[i]
        D[delta[ri] + w[ri]] = i
        w[ri] += 1


@cython.boundscheck(False)
cpdef _make_donors_to_n(DTYPE_INT_t np,
                  DTYPE_INT_t q,
                  np.ndarray[DTYPE_INT_t, ndim=1] w,
                  np.ndarray[DTYPE_INT_t, ndim=1] D,
                  np.ndarray[DTYPE_INT_t, ndim=1] delta,
                  np.ndarray[DTYPE_INT_t, ndim=2] r,
                  np.ndarray[DTYPE_FLOAT_t, ndim=2] p,
                  ):
    """Determines number of donors for route to n"""
    cdef int ri, i, v, ind
    for v in range(q):
        for i in range(np):
            ri = r[i, v]
            if p[i, v] > 0:
                ind = delta[ri] + w[ri]
                D[ind] = i
                w[ri] += 1
